from copy import deepcopy
from typing import Any, Dict, Tuple, DefaultDict, Optional, List
from collections import defaultdict
from urllib.parse import urlparse
from dataclasses import dataclass, field

from retrying import retry  # type: ignore

from capella_console_client.logconf import logger
from capella_console_client.session import CapellaConsoleSession
from capella_console_client.config import (
    ALL_SUPPORTED_FIELDS,
    ALL_SUPPORTED_SORTBY,
    SUPPORTED_SEARCH_FIELDS,
    SUPPORTED_QUERY_FIELDS,
    STAC_PREFIXED_BY_QUERY_FIELDS,
    OPERATOR_SUFFIXES,
    DEFAULT_PAGE_SIZE,
    DEFAULT_MAX_FEATURE_COUNT,
)
from capella_console_client.hooks import retry_if_http_status_error, log_attempt_delay


@dataclass
class SearchResult:

    request_body: Dict[str, Any] = field(default_factory=dict)
    _pages: List[Dict[str, Any]] = field(default_factory=list)
    _features: List[Dict[str, Any]] = field(default_factory=list)

    def add(self, page: Dict[str, Any], keep_duplicates: bool = False) -> int:
        if not keep_duplicates:
            page = self._filter_dupes(page)
        self._pages.append(page)
        self._features.extend(page["features"])
        return len(page["features"])

    def _filter_dupes(self, page: Dict[str, Any]) -> Dict[str, Any]:
        # drop duplicates within page features
        page_stac_ids = [feat["id"] for feat in page["features"]]
        set_page_stac_ids = set(page_stac_ids)

        if len(set_page_stac_ids) != len(page["features"]):
            page["features"] = [page["features"][page_stac_ids.index(p_id)] for p_id in set_page_stac_ids]

        # drop duplicates within SearchResult
        dupes = set_page_stac_ids.intersection(self.stac_ids)

        if dupes:
            page["features"] = [p for p in page["features"] if p["id"] not in dupes]

        return page

    # backwards compability
    def __getitem__(self, key):
        return self._features.__getitem__(key)

    def __iter__(self):
        return self._features.__iter__()

    def __len__(self):
        return len(self._features)

    def __repr__(self):
        return f"{self.__class__} ({len(self)} STAC items)"

    def to_feature_collection(self):
        return {"type": "FeatureCollection", "features": self._features}

    @property
    def stac_ids(self):
        return [item["id"] for item in self._features]

    def merge(self, other: "SearchResult", keep_duplicates: bool = False) -> "SearchResult":
        copy = deepcopy(self)
        for page in other._pages:
            copy.add(page=page, keep_duplicates=keep_duplicates)

        return copy


class StacSearch:
    def __init__(self, session: CapellaConsoleSession, **kwargs) -> None:
        cur_kwargs = deepcopy(kwargs)
        self.session = session
        self.payload: Dict[str, Any] = {}

        sortby = cur_kwargs.pop("sortby", None)
        query_payload = self._get_query_payload(cur_kwargs)
        if query_payload:
            self.payload["query"] = dict(query_payload)

        if sortby:
            self.payload["sortby"] = self._get_sort_payload(sortby)

    def _get_query_payload(self, kwargs) -> DefaultDict[str, Dict[str, Any]]:
        query_payload: DefaultDict[str, Dict[str, Any]] = defaultdict(dict)

        for cur_field, value in kwargs.items():
            cur_field, op = self._split_op(cur_field)
            if cur_field not in ALL_SUPPORTED_FIELDS:
                logger.warning(f"filter {cur_field} not supported ... omitting")
                continue

            if op not in OPERATOR_SUFFIXES:
                logger.warning(f"operator {op} not supported ... omitting")
                continue

            if cur_field in SUPPORTED_SEARCH_FIELDS:
                self.payload[cur_field] = value
            elif cur_field in SUPPORTED_QUERY_FIELDS:
                if type(value) == list:
                    op = "in"

                target_field = STAC_PREFIXED_BY_QUERY_FIELDS.get(cur_field, cur_field)
                query_payload[target_field][op] = value

        return query_payload

    def _split_op(self, cur_field: str) -> Tuple[str, str]:
        parts = cur_field.split("__")
        if len(parts) == 2:
            op = parts[1]
        else:
            op = "eq"
        return (parts[0], op)

    def _get_sort_payload(self, sortby):
        directions = {"-": "desc", "+": "asc"}
        sorts = []
        orig = sortby

        if not isinstance(orig, list):
            orig = [orig]

        for sort_arg in orig:
            field = sort_arg[1:]
            direction = sort_arg[0]
            if direction not in directions:
                direction = "+"
                field = sort_arg

            if field not in ALL_SUPPORTED_SORTBY:
                logger.warning(f"sorting by {field} not supported ... omitting")
                continue

            if field in SUPPORTED_QUERY_FIELDS or field == "datetime":
                field = f"properties.{field}"

            sorts.append({"field": field, "direction": directions[direction]})
        return sorts

    def fetch_all(self) -> SearchResult:
        logger.info(f"searching catalog with payload {self.payload}")

        requested_limit = self.payload.get("limit", DEFAULT_MAX_FEATURE_COUNT)

        if "limit" not in self.payload:
            self.payload["limit"] = DEFAULT_MAX_FEATURE_COUNT

        # ensure DEFAULT_PAGE_SIZE if requested limit > DEFAULT_PAGE_SIZE
        self.payload["limit"] = min(DEFAULT_PAGE_SIZE, self.payload["limit"])

        page_cnt = 1
        search_result = SearchResult(request_body=self.payload)
        next_href = None

        while True:
            _log_page_query(page_cnt, len(search_result), self.payload["limit"])
            page_data = _page_search(self.session, self.payload, next_href)
            number_matched = page_data["numberMatched"]
            items_added = search_result.add(page_data)

            limit_reached = len(search_result) >= requested_limit or len(search_result) >= number_matched

            # all dupes
            size_unchanged = items_added == 0
            if limit_reached or size_unchanged:
                break

            next_href = _get_next_page_href(page_data)
            if next_href is None:
                break

            if page_cnt == 1:
                logger.info(f"Matched a total of {number_matched} stac items")

            self.payload["limit"] = DEFAULT_PAGE_SIZE
            page_cnt += 1
            self.payload["page"] = page_cnt

        # truncate to limit
        len_features = len(search_result)
        if len_features > requested_limit:
            search_result._features = search_result._features[:requested_limit]

        if not len_features:
            logger.info("found no STAC items matching your query")
        else:
            multiple_suffix = "s" if len_features > 1 else ""
            logger.info(f"found {len(search_result)} STAC item{multiple_suffix}")

        return search_result


def _log_page_query(page_cnt: int, len_feat: int, limit: int):
    if page_cnt != 1:
        logger.info(f"\tpage {page_cnt} ({len_feat} - {len_feat + limit})")


def _get_next_page_href(page_data: Dict[str, Any]) -> Optional[str]:
    links = page_data.get("links", [])
    try:
        next_href: Optional[str] = next(filter(lambda c: c["rel"] == "next", links))["href"]
    except StopIteration:
        next_href = None

    return next_href


@retry(
    retry_on_exception=retry_if_http_status_error,
    wait_func=log_attempt_delay,
    wait_exponential_multiplier=1000,
    stop_max_delay=16000,
)
def _page_search(session: CapellaConsoleSession, payload: Dict[str, Any], next_href: str = None) -> Dict[str, Any]:
    if next_href:
        # STAC API to return normalized asset hrefs, not api gateway - fixing this here ...
        url_parsed = urlparse(next_href)
        if url_parsed.netloc != urlparse(session.search_url).netloc:
            next_href = f"{session.search_url}?{url_parsed.query}"

    url = session.search_url if next_href is None else next_href
    resp = session.post(url, json=payload)

    data: Dict[str, Any] = resp.json()
    return data



from enum import Enum, EnumMeta


class MetaEnum(EnumMeta):
    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        return True


class BaseEnum(Enum, metaclass=MetaEnum):
    pass


class ProductType(str, BaseEnum):
    SLC = "SLC"
    GEO = "GEO"
    SICD = "SICD"
    GEC = "GEC"
    SIDD = "SIDD"
    CPHD = "CPHD"


class AssetType(str, BaseEnum):
    HH = "HH"
    VV = "VV"
    preview = "preview"
    raster = "raster"
    metadata = "metadata"
    thumbnail = "thumbnail"
    log = "log"
    profile = "profile"
    stats = "stats"
    stats_plots = "stats_plots"


class ProductClass(str, BaseEnum):
    standard = "standard"
    extended = "extended"
    custom = "custom"


class OrbitState(str, BaseEnum):
    ascending = "ascending"
    descending = "descending"


class ObservationDirection(str, BaseEnum):
    left = "left"
    right = "right"


class OrbitalPlane(str, BaseEnum):
    fortyfive = 45
    fiftythree = 53
    ninetyseven = 97


class InstrumentMode(str, BaseEnum):
    stripmap = "stripmap"
    spotlight = "spotlight"
    sliding_spotlight = "sliding_spotlight"



import typer
from tabulate import tabulate
import questionary

from capella_console_client.cli.cache import CLICache
from capella_console_client.cli.info import no_data_info
from capella_console_client.cli.user_searches.core import (
    rename_search_entity,
    SearchEntity,
    _load_and_prompt,
)
from capella_console_client.logconf import logger


app = typer.Typer(help="manage saved search results")


@app.command()
def list(detailed: bool = typer.Option(False, "--detailed", help="Show STAC ids of saved search results")):
    """
    list previously saved search results
    """
    saved_search_results = CLICache.load_my_search_results()
    if not saved_search_results:
        no_data_info(search_entity=SearchEntity.result)

    table_data = []
    for k, v in saved_search_results.items():
        stac_ids = v["data"]
        cur = [k, v["created_at"], v["updated_at"], f"{len(stac_ids)} items"]
        if detailed:
            cur.append("\n".join(stac_ids))
        table_data.append(cur)

    headers = ["identifier", "created", "updated", "size", "STAC ids"]
    logger.info(typer.style("My saved search results\n", bold=True))
    typer.secho(tabulate(table_data, tablefmt="fancy_grid", headers=headers))
    return saved_search_results


@app.command()
def rename():
    """
    rename previously saved search result
    """
    rename_search_entity(search_entity=SearchEntity.result)


@app.command()
def delete():
    """
    delete previously saved search result
    """
    saved_search_results, selection = _load_and_prompt(
        "Which saved search result would you like to delete?", SearchEntity.result
    )

    if not selection:
        return

    for selected in selection:
        del saved_search_results[selected]

    CLICache.write_my_search_results(saved_search_results)
    suffix = "s" if len(selection) > 1 else ""
    logger.info(f"Deleted the following search result{suffix}:")
    logger.info("\n".join(selected))


@app.command()
def prune():
    """
    delete ALL previously saved search result
    """
    if questionary.confirm(
        "Please confirm you'd like to delete ALL of your saved search results. This action cannot be undone."
    ).ask():
        try:
            CLICache.MY_SEARCH_RESULTS.unlink()
        except FileNotFoundError:
            pass
        logger.info(f"Deleted ALL saved search results")



from typing import Dict, Any


class CapellaConsoleClientError(Exception):
    response = None
    data: Dict[str, Any] = {}
    message = "An unknown error occurred"

    def __init__(self, message=None, code=None, data={}, response=None):
        self.response = response
        if message:
            self.message = message
        if code:
            self.code = code
        if data:
            self.data = data


class AuthenticationError(CapellaConsoleClientError):
    pass


class NoValidStacIdsError(CapellaConsoleClientError):
    pass


class TaskNotCompleteError(CapellaConsoleClientError):
    pass


class InsufficientFundsError(CapellaConsoleClientError):
    pass


class OrderRejectedError(CapellaConsoleClientError):
    pass


class OrderExpiredError(CapellaConsoleClientError):
    pass


class ConnectError(CapellaConsoleClientError):
    pass


class CollectionAccessDeniedError(CapellaConsoleClientError):
    pass


class AuthorizationError(CapellaConsoleClientError):
    pass


class NoRefreshTokenError(CapellaConsoleClientError):
    pass


DEFAULT_ERROR_CODE = "GENERAL_API_ERROR"
INVALID_TOKEN_ERROR_CODE = "INVALID_TOKEN"
ORDER_EXPIRED_ERROR_CODE = "ORDER_EXPIRED"
COLLECTION_ACCESS_DENIED_ERROR_CODE = "COLLECTION_ACCESS_DENIED"
NOT_AUTHORIZED_ERROR_CODE = "NOT_AUTHORIZED"

ERROR_CODES = {
    INVALID_TOKEN_ERROR_CODE: AuthenticationError,
    ORDER_EXPIRED_ERROR_CODE: OrderExpiredError,
    COLLECTION_ACCESS_DENIED_ERROR_CODE: CollectionAccessDeniedError,
    NOT_AUTHORIZED_ERROR_CODE: AuthorizationError,
}

ERROR_CODES_BY_MESSAGE_SNIP = {
    "order expired": ORDER_EXPIRED_ERROR_CODE,
    "not permitted to access any of the collection": COLLECTION_ACCESS_DENIED_ERROR_CODE,
    "not authorized to perform": NOT_AUTHORIZED_ERROR_CODE,
}

NON_RETRYABLE_ERROR_CODES = (
    INVALID_TOKEN_ERROR_CODE,
    COLLECTION_ACCESS_DENIED_ERROR_CODE,
    NOT_AUTHORIZED_ERROR_CODE,
)


def handle_error_response(resp):
    error = resp.json()
    try:
        if "error" in error:
            error = error["error"]

        message = error.get("message", error.get("Message"))
        code = error.get("code", DEFAULT_ERROR_CODE)
        data = error.get("data", {})
    except Exception:
        message = error
        code = DEFAULT_ERROR_CODE
        data = {}

    # try to assign some more meaningful exception class by message
    if code == DEFAULT_ERROR_CODE and message:
        try:
            code = next(v for k, v in ERROR_CODES_BY_MESSAGE_SNIP.items() if k in message.lower())
        except StopIteration:
            pass

    exc = ERROR_CODES.get(code, CapellaConsoleClientError)(message=message, code=code, data=data, response=resp)
    raise exc



from typing import List

import typer
import questionary

from tabulate import tabulate

from capella_console_client.cli.validate import (
    _must_be_type,
    _validate_dir_exists,
    _validate_email,
    _no_selection_bye,
    _at_least_one_selected,
)
from capella_console_client.cli.cache import CLICache
from capella_console_client.cli.config import (
    CLI_SUPPORTED_RESULT_HEADERS,
    CURRENT_SETTINGS,
    SearchFilterOrderOption,
)
from capella_console_client.cli.prompt_helpers import get_first_checked
from capella_console_client.logconf import logger


app = typer.Typer(help="fine tune settings")


def _prompt_search_result_headers() -> List[str]:
    choices = [
        questionary.Choice(cur, checked=cur in CURRENT_SETTINGS["search_headers"])
        for cur in CLI_SUPPORTED_RESULT_HEADERS
    ]

    search_result_fields = questionary.checkbox(
        "Which STAC item fields would you like to display in the search results table ?",
        choices=choices,
        initial_choice=get_first_checked(choices),
        validate=_at_least_one_selected,
    ).ask()
    _no_selection_bye(search_result_fields, info_msg="no valid path provided")

    return search_result_fields


@app.command()
def show():
    """
    show current settings
    """
    typer.secho("Current settings:\n", underline=True)
    table_data = list(CURRENT_SETTINGS.items())
    typer.echo(tabulate(table_data, tablefmt="fancy_grid", headers=["setting", "value"]))


@app.command()
def result_table():
    """
    set fields (STAC properties) of search results table
    """
    search_result_headers = _prompt_search_result_headers()

    CLICache.write_user_settings("search_headers", search_result_headers)
    typer.echo("updated fields that to will be displayed in search results table")


@app.command()
def limit():
    """
    set default limit to be used in searches
    """
    limit = questionary.text(
        "Specify default limit to be used in searches (can be overridden at search time):",
        default=str(CURRENT_SETTINGS["limit"]),
        validate=_must_be_type(int),
    ).ask()
    _no_selection_bye(limit, info_msg="no valid limit provided")

    if int(limit) > 0:
        CLICache.write_user_settings("limit", int(limit))
        typer.echo(f"updated default search limit to {limit}")
    else:
        typer.echo("invalid limit")


@app.command()
def user():
    """
    set user for Capella Console
    """
    console_user = questionary.path(
        "User on console.capellaspace.com (user@email.com):",
        default=CURRENT_SETTINGS.get("console_user", ""),
        validate=_validate_email,
    ).ask()
    _no_selection_bye(console_user, info_msg="no user provided")

    CLICache.write_user_settings("console_user", console_user)
    typer.echo("updated user for Capella Console")
    CLICache.JWT.unlink(missing_ok=True)


@app.command()
def output():
    """
    set default output location for downloads and .json STAC exports
    """
    out_path = questionary.path(
        "Specify the default location for downloads and .json STAC exports: (press <tab>)",
        default=CURRENT_SETTINGS["out_path"],
        validate=_validate_dir_exists,
    ).ask()
    _no_selection_bye(out_path)

    CLICache.write_user_settings("out_path", out_path)
    typer.echo("updated default output path for .json STAC exports")


@app.command()
def search_filter_order():
    """
    set order of search filters to be used in searches
    """
    search_filter_order = questionary.select(
        "Specify the order of search filters to be used in searches:",
        choices=list(SearchFilterOrderOption),
        default=SearchFilterOrderOption[CURRENT_SETTINGS["search_filter_order"]],
    ).ask()

    _no_selection_bye(search_filter_order, info_msg="no valid search filter order provided")
    CLICache.write_user_settings("search_filter_order", SearchFilterOrderOption(search_filter_order).name)
    typer.echo("updated order of search filters to be used in searches")


def configure():
    logger.info(typer.style("let's get you all setup using capella-console-wizard:", bold=True))
    logger.info("\t\tPress Ctrl + C anytime to quit\n")
    user()
    output()
    search_filter_order()
    result_table()
    limit()



from copy import deepcopy
from pathlib import Path
from typing import Dict, Any

from capella_console_client.cli.cache import CLICache
from capella_console_client.config import ALL_SUPPORTED_FIELDS
from capella_console_client.enumerations import (
    InstrumentMode,
    ProductClass,
    ObservationDirection,
    OrbitState,
    OrbitalPlane,
    ProductType,
    BaseEnum,
)


class SearchFilterOrderOption(str, BaseEnum):
    console_ui = "console UI filters on top"
    alphabetical = "alphabetical"
    # TODO
    # custom = "custom"


DEFAULT_SEARCH_RESULT_HEADERS = [
    "id",
    "instrument_mode",
    "product_type",
    "polarizations",
    "incidence_angle",
]

DEFAULT_SETTINGS = {
    "limit": 50,
    "search_headers": DEFAULT_SEARCH_RESULT_HEADERS,
    "out_path": str(Path.home()),
    "order_list_limit": 50,
    "search_filter_order": SearchFilterOrderOption.console_ui.name,
}


USER_SETTINGS = CLICache.load_user_settings()

CURRENT_SETTINGS: Dict[str, Any] = {**DEFAULT_SETTINGS, **USER_SETTINGS}


CLI_SEARCH_FIELDS = [
    "instrument_mode",
    "product_type",
    "bbox",
    "datetime",
    "orbit_state",
    "observation_direction",
    "incidence_angle",
    "resolution_range",
    "resolution_azimuth",
    "product_category",
    "collections",
    "ids",
    "limit",
    "center_frequency",
    "collect_id",
    "constellation",
    "frequency_band",
    "instruments",
    "look_angle",
    "looks_azimuth",
    "looks_equivalent_number",
    "looks_range",
    "pixel_spacing_azimuth",
    "pixel_spacing_range",
    "platform",
    "polarizations",
    "resolution_ground_range",
    "squint_angle",
    "orbital_plane",
    "billable_area",
]


def get_cli_supported_search_filters():
    if CURRENT_SETTINGS["search_filter_order"] == SearchFilterOrderOption.alphabetical:
        return sorted(CLI_SEARCH_FIELDS)
    return CLI_SEARCH_FIELDS


CLI_SUPPORTED_SEARCH_FILTERS = get_cli_supported_search_filters()

CLI_SUPPORTED_RESULT_HEADERS = deepcopy(CLI_SUPPORTED_SEARCH_FILTERS)
CLI_SUPPORTED_RESULT_HEADERS[CLI_SUPPORTED_RESULT_HEADERS.index("ids")] = "id"
CLI_SUPPORTED_RESULT_HEADERS[CLI_SUPPORTED_RESULT_HEADERS.index("collections")] = "collection"

PROMPT_OPERATORS = {
    "billable_area",
    "center_frequency",
    "datetime",
    "incidence_angle",
    "look_angle",
    "looks_azimuth",
    "looks_equivalent_number",
    "looks_range",
    "pixel_spacing_azimuth",
    "pixel_spacing_range",
    "resolution_azimuth",
    "resolution_ground_range",
    "resolution_range",
    "squint_angle",
}


ENUM_CHOICES_BY_FIELD_NAME = {
    "instrument_mode": InstrumentMode,
    "observation_direction": ObservationDirection,
    "orbital_plane": OrbitalPlane,
    "orbit_state": OrbitState,
    "product_category": ProductClass,
    "product_type": ProductType,
}

COLLECTIONS = [
    "capella-cphd",
    "capella-gec",
    "capella-geo",
    "capella-open-data",
    "capella-sicd",
    "capella-sidd",
    "capella-slc",
    "other",
]



"""

"""

import sys


import typer

from capella_console_client.exceptions import AuthenticationError
from capella_console_client.cli.client_singleton import CLIENT
from capella_console_client.cli.config import CURRENT_SETTINGS
import capella_console_client.cli.settings
import capella_console_client.cli.user_searches.my_searches
import capella_console_client.cli.orders
import capella_console_client.cli.checkout
import capella_console_client.cli.workflows
from capella_console_client.cli.cache import CLICache
from capella_console_client.logconf import logger


def auto_auth_callback(ctx: typer.Context):
    # TODO: how to do this properly
    RETURN_EARLY_ON = ("--help",)
    if any(ret in sys.argv for ret in RETURN_EARLY_ON):
        return

    if ctx.invoked_subcommand in (
        "settings",
        "my-search-results",
        "my-search-queries",
        None,
    ):
        return

    if ctx.invoked_subcommand == sys.argv[-1]:
        return

    try:
        CLIENT._sesh.authenticate(token=CLICache.load_jwt(), no_token_check=False)
    # first time or expired token
    except (FileNotFoundError, AuthenticationError):
        auth_kwargs = {}
        if "console_user" in CURRENT_SETTINGS:
            auth_kwargs["email"] = CURRENT_SETTINGS["console_user"]
            logger.info(f"authenticating as {auth_kwargs['email']}")
        CLIENT._sesh.authenticate(**auth_kwargs)  # type: ignore
        jwt = CLIENT._sesh.headers["authorization"]
        CLICache.write_jwt(jwt)


app = typer.Typer(
    help="Interactive wizard for api.capellaspace.com",
    callback=auto_auth_callback,
)
app.add_typer(capella_console_client.cli.settings.app, name="settings")
app.add_typer(
    capella_console_client.cli.user_searches.my_searches.app,
    name="my-searches",
)
app.add_typer(capella_console_client.cli.orders.app, name="orders")
app.add_typer(capella_console_client.cli.workflows.app, name="workflows")


@app.command()
def configure():
    """
    configure capella-console-wizard
    """
    capella_console_client.cli.settings.configure()


def main():
    app()


if __name__ == "__main__":
    main()



import json

import typer
from tabulate import tabulate
import questionary

from capella_console_client.cli.cache import CLICache
from capella_console_client.cli.info import no_data_info
from capella_console_client.cli.user_searches.core import (
    rename_search_entity,
    SearchEntity,
    _load_and_prompt,
)

app = typer.Typer(help="manage saved search queries")


@app.command()
def list(detailed: bool = typer.Option(False, "--detailed", help="Show STAC ids of saved search queries")):
    """
    list previously saved search queries
    """
    saved_search_queries = CLICache.load_my_search_queries()
    if not saved_search_queries:
        no_data_info(search_entity=SearchEntity.query)

    table_data = []
    for k, v in saved_search_queries.items():
        data = v["data"]
        cur = [
            k,
            v["created_at"],
            v["updated_at"],
            json.dumps(data, sort_keys=True, indent=4),
        ]
        table_data.append(cur)

    typer.secho("My saved search queries\n", bold=True)
    typer.echo(
        tabulate(
            table_data,
            tablefmt="fancy_grid",
            headers=["identifier", "created at", "updated at", "query"],
        )
    )
    return saved_search_queries


@app.command()
def rename():
    """
    rename previously saved search result
    """
    rename_search_entity(search_entity=SearchEntity.query)


@app.command()
def delete():
    """
    delete previously saved search query
    """
    saved_search_queries, selection = _load_and_prompt(
        "Which saved search queries would you like to delete?",
        search_entity=SearchEntity.query,
    )

    if not selection:
        return

    for selected in selection:
        del saved_search_queries[selected]

    CLICache.write_my_search_queries(saved_search_queries)
    typer.echo(f"Deleted {len(selection)} search queries")


@app.command()
def prune():
    """
    delete ALL previously saved search queries
    """
    if questionary.confirm(
        "Please confirm you'd like to delete ALL of your saved search queries. This action cannot be undone."
    ).ask():
        try:
            CLICache.MY_SEARCH_QUERIES.unlink()
        except FileNotFoundError:
            pass
        typer.echo(f"Deleted ALL saved search queries")



import typer

from capella_console_client.cli.search import interactive_search, from_saved
from capella_console_client.cli.checkout import interactive_search_order_and_download

app = typer.Typer(help="interactive workflows")


@app.command(help="interactively search through the archive")
def search():
    interactive_search()


@app.command(help="re-search a previously saved query")
def search_from_saved():
    from_saved()


@app.command(help="interactively search, order and download workflow")
def checkout():
    interactive_search_order_and_download()



import json
from pathlib import Path
from typing import List, Dict, Any, Union
from datetime import datetime

from capella_console_client.logconf import logger


def _safe_load_json(file_path: Path) -> Dict[str, Any]:
    content = {}
    try:
        content = json.loads(file_path.read_text())
    except:
        pass
    return content


class CLICache:
    ROOT = Path.home() / ".capella-console-wizard"
    JWT = ROOT / "jwt.cache"
    SETTINGS = ROOT / "settings.json"
    MY_SEARCH_RESULTS = ROOT / "my-search-results.json"
    MY_SEARCH_QUERIES = ROOT / "my-search-queries.json"

    @classmethod
    def write_jwt(cls, jwt: str):
        cls.JWT.write_text(jwt)
        logger.info(f"Cached JWT to {cls.JWT}")

    @classmethod
    def load_jwt(cls) -> str:
        return cls.JWT.read_text()

    @classmethod
    def write_user_settings(cls, key: str, value: Any):
        settings = cls.load_user_settings()
        settings[key] = value
        cls.SETTINGS.write_text(json.dumps(settings))

    @classmethod
    def load_user_settings(cls) -> Dict[str, Any]:
        return _safe_load_json(cls.SETTINGS)

    @classmethod
    def add_timestamps(cls, data: Union[Dict[str, Any], List[str]], is_new: bool = False) -> Dict[str, Any]:
        now = str(datetime.now())[:-7]
        record = {
            "data": data,
            "updated_at": now,
        }
        if is_new:
            record["created_at"] = now
        return record

    @classmethod
    def write_my_search_results(cls, my_search_results: Dict[str, Any]):
        cls.MY_SEARCH_RESULTS.write_text(json.dumps(my_search_results))

    @classmethod
    def update_my_search_results(cls, search_identifier: str, stac_ids: List[str], is_new: bool = False):
        my_search_results = cls.load_my_search_results()
        my_search_results[search_identifier] = cls.add_timestamps(stac_ids, is_new)
        cls.write_my_search_results(my_search_results)

    @classmethod
    def load_my_search_results(cls) -> Dict[str, Any]:
        return _safe_load_json(cls.MY_SEARCH_RESULTS)

    @classmethod
    def write_my_search_queries(cls, my_queries: Dict[str, Any]):
        cls.MY_SEARCH_QUERIES.write_text(json.dumps(my_queries))

    @classmethod
    def update_my_search_queries(cls, search_identifier: str, search_query: Dict[str, Any], is_new: bool = False):
        my_queries = cls.load_my_search_queries()
        my_queries[search_identifier] = cls.add_timestamps(search_query, is_new)
        cls.write_my_search_queries(my_queries)

    @classmethod
    def load_my_search_queries(cls) -> Dict[str, Any]:
        return _safe_load_json(cls.MY_SEARCH_QUERIES)


CLICache.ROOT.mkdir(exist_ok=True)



import typer
import questionary
from questionary import prompt

from capella_console_client.enumerations import ProductType
from capella_console_client.cli.user_searches.core import _load_and_prompt, SearchEntity
from capella_console_client.cli.config import CURRENT_SETTINGS
from capella_console_client.cli.client_singleton import CLIENT
from capella_console_client.cli.validate import _validate_uuid, _validate_dir_exists
from capella_console_client.enumerations import BaseEnum
from capella_console_client.cli.search import (
    _prompt_search_filters,
    search_and_post_actions,
)
from capella_console_client.cli.orders import (
    _list_orders_and_tabulate,
    PostOrderListActions,
)


app = typer.Typer(help="order and download products")


class CheckoutStartOptions(str, BaseEnum):
    new_search = "new search"
    saved_search = "use previously saved search results"
    collect_id = "provide a collect id"
    tasking_request_id = "provide a taskingrequest id"
    existing_order = "select existing order"

    @classmethod
    def _get_choices(cls):
        return list(CheckoutStartOptions)


def interactive_search_order_and_download():

    start_from_opt = questionary.select("What would you like to do?", choices=CheckoutStartOptions._get_choices()).ask()

    questions = _get_questions(start_from_opt)
    if start_from_opt in [
        CheckoutStartOptions.collect_id,
        CheckoutStartOptions.tasking_request_id,
    ]:
        answers = prompt(questions)
        if answers["include"] == "all":
            del answers["include"]
        paths = CLIENT.download_products(**answers)

    elif start_from_opt in (
        CheckoutStartOptions.new_search,
        CheckoutStartOptions.saved_search,
    ):
        if start_from_opt == CheckoutStartOptions.new_search:
            search_query = _prompt_search_filters()
            stac_items = search_and_post_actions(search_query)
            answers = prompt(questions)
            order_id = CLIENT.submit_order(
                items=stac_items,
                check_active_orders=True,
                omit_search=True,
            )
            stac_ids = [s["id"] for s in stac_items]

        else:
            stac_ids = _stac_ids_from_saved_search()
            answers = prompt(questions)
            order_id = CLIENT.submit_order(
                stac_ids=stac_ids,
                check_active_orders=True,
                omit_search=True,
            )

        assets_presigned = CLIENT.get_presigned_assets(order_id, stac_ids=stac_ids)
        paths = CLIENT.download_products(assets_presigned=assets_presigned, **answers)
    elif start_from_opt == CheckoutStartOptions.existing_order:
        orders = _list_orders_and_tabulate(is_active=False, limit=300)
        order_id = PostOrderListActions.prompt_and_reorder(orders)
        answers = prompt(questions)
        paths = CLIENT.download_products(**answers, order_id=order_id)

    product_paths = set()
    for stac_id in paths:
        asset_paths = list(paths[stac_id].values())
        product_paths.update([a.parent for a in asset_paths])

    if questionary.confirm("Do you want to open any product directories?").ask():
        dirs_to_open = questionary.checkbox(
            "select which product directories you want to open",
            choices=list(map(str, product_paths)),
        ).ask()

        for open_dir in dirs_to_open:
            typer.launch(open_dir)


def _get_questions(start_option: CheckoutStartOptions):
    question_types = {
        "uuid": {
            "type": "text",
            "name": start_option.name,
            "message": f"{start_option.value}:",
            "validate": _validate_uuid,
        },
        "product_types": {
            "type": "checkbox",
            "name": "product_types",
            "message": "product type(s):",
            "choices": list(ProductType),
        },
        "asset_types": {
            "type": "checkbox",
            "name": "include",
            "message": "asset type:",
            "choices": ["all", "raster", "metadata", "thumbnail"],
        },
        "local_dir": {
            "type": "path",
            "name": "local_dir",
            "message": "download location:",
            "default": CURRENT_SETTINGS["out_path"],
            "validate": _validate_dir_exists,
        },
    }

    return {
        CheckoutStartOptions.collect_id: [
            question_types["uuid"],
            question_types["product_types"],
            question_types["asset_types"],
            question_types["local_dir"],
        ],
        CheckoutStartOptions.tasking_request_id: [
            question_types["uuid"],
            question_types["product_types"],
            question_types["asset_types"],
            question_types["local_dir"],
        ],
        CheckoutStartOptions.existing_order: [
            question_types["asset_types"],
            question_types["local_dir"],
        ],
        CheckoutStartOptions.new_search: [
            question_types["asset_types"],
            question_types["local_dir"],
        ],
        CheckoutStartOptions.saved_search: [
            question_types["asset_types"],
            question_types["local_dir"],
        ],
    }[start_option]


def _stac_ids_from_saved_search():
    saved, selection = _load_and_prompt(
        "Which saved search would you like to use?",
        search_entity=SearchEntity.result,
        multiple=False,
    )
    stac_ids = saved[selection]["data"]
    return stac_ids



from typing import List, Dict, Any, Optional
from collections import defaultdict

import typer
from tabulate import tabulate

from capella_console_client.config import (
    STAC_PREFIXED_BY_QUERY_FIELDS,
)
from capella_console_client.search import SearchResult

from capella_console_client.cli.config import (
    CURRENT_SETTINGS,
)


def show_tabulated(
    stac_items: SearchResult,
    search_headers: Optional[List[str]] = None,
    show_row_number: bool = False,
):
    if not search_headers:
        search_headers = CURRENT_SETTINGS["search_headers"]  # type: ignore

    table_data = defaultdict(list)

    assert search_headers is not None
    # force id left if specified
    if "id" in search_headers:
        del search_headers[search_headers.index("id")]
        search_headers.insert(0, "id")

    if show_row_number:
        table_data["#"] = list(range(1, len(stac_items) + 1))

    for field in search_headers:
        for it in stac_items:
            if field in STAC_PREFIXED_BY_QUERY_FIELDS:
                value = it["properties"].get(STAC_PREFIXED_BY_QUERY_FIELDS[field])
            else:
                value = it["properties"].get(field)
                if value is None:
                    value = it.get(field, "n/a")

            if value is not None:
                table_data[field].append(value)

    typer.echo(tabulate(table_data, tablefmt="fancy_grid", headers="keys"))
    typer.echo("\n")


def show_orders_tabulated(orders: List[Dict[str, Any]]):
    fields = ["orderId", "orderDate", "expirationDate", "orderStatus"]
    table_data = []
    for i, order in enumerate(orders):
        cur = [i + 1]
        cur.extend(order[field] for field in fields)
        granules = [o["granuleId"] for o in order["items"]]
        cur.append("\n".join(granules))  # type: ignore
        table_data.append(cur)

    headers = ["#", *fields, "STAC ids"]
    typer.echo(tabulate(table_data, tablefmt="fancy_grid", headers=headers))


def show_order_review_tabulated(order_review: Dict[str, Any]):
    order_summary = order_review["orderDetails"]["summary"]
    typer.secho("summary:\n", bold=True)
    table_data = [
        ("field", "value"),
        ("authorized", order_review["authorized"]),
        ("subtotal", order_summary["subtotal"]),
        ("total", order_summary["total"]),
    ]
    typer.echo(tabulate(table_data, tablefmt="fancy_grid", headers="firstrow"))

    typer.secho("\n\nby_line_item:\n", bold=True)
    line_items = order_review["orderDetails"]["lineItems"]

    line_item_table = []
    for li in line_items:
        cur = (li["granuleId"], li["order"]["finalListPrice"])
        line_item_table.append(cur)

    typer.echo(tabulate(line_item_table, tablefmt="fancy_grid", headers=["STAC id", "cost"]))



import json
from uuid import UUID
from pathlib import Path
from typing import List, Dict, Any, Iterable


class SafeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, UUID):
            # if the obj is uuid, we simply return the value of uuid
            return obj.hex
        return json.JSONEncoder.default(self, obj)


def convert_to_uuid_str(dict_args: Dict[str, Any], uuid_arg_names: Iterable[str]) -> Dict[str, Any]:
    for conv in uuid_arg_names:
        if dict_args.get(conv):
            dict_args[conv] = str(dict_args[conv])

    return dict_args



import uuid
import re
from collections import Counter

from typing import no_type_check, Optional, List, Dict, Any, Union

from capella_console_client.enumerations import ProductType, AssetType
from capella_console_client.logconf import logger
from capella_console_client.search import SearchResult

STAC_ID_REGEX_STRICT = re.compile("^CAPELLA_C\\d{2}_\\w+_\\w+_\\w{2}_\\d{14}_\\d{14}$")


@no_type_check
def _validate_uuid(uuid_str: str) -> None:
    try:
        uuid.UUID(uuid_str)
    except ValueError as e:
        raise ValueError(f"{uuid_str} is not a valid uuid: {e}")


def _validate_stac_id_or_stac_items(
    stac_ids: Optional[List[str]] = None,
    items: Union[Optional[List[Dict[str, Any]]], SearchResult] = None,
) -> List[str]:
    if not stac_ids and not items:
        raise ValueError("Please provide stac_ids or items")

    if not stac_ids:
        stac_ids = [f["id"] for f in items]  # type: ignore

    return stac_ids


def _validate_and_filter_product_types(
    product_types: Optional[List[str]],
) -> Optional[List[str]]:
    if not product_types:
        return None
    return [p.upper() for p in product_types if p.upper() in ProductType]


def _validate_and_filter_asset_types(
    asset_types: Union[List[str], str, None],
) -> Optional[List[str]]:
    if not asset_types:
        return None
    if isinstance(asset_types, str):
        return [a for a in [asset_types] if a in AssetType]
    return [a for a in asset_types if a in AssetType]


def _validate_and_filter_stac_ids(stac_ids: Optional[List[str]]) -> List[str]:
    if not stac_ids:
        return []

    valid_stac_ids = list(set(filter(STAC_ID_REGEX_STRICT.match, stac_ids)))

    diff = set(stac_ids) - set(valid_stac_ids)
    if diff:
        logger.warning(f"filtered {','.join(diff)} (no valid STAC id)")

    if not valid_stac_ids:
        logger.warning("No valid STAC id provided")
        return []

    duplicates = [k for k, v in Counter(stac_ids).items() if v > 1]
    if duplicates:
        logger.warning(f"filtered {','.join(duplicates)} (duplicate)")

    return valid_stac_ids



from pathlib import Path
from urllib.parse import urlparse
from dataclasses import dataclass
import tempfile
from concurrent.futures import ThreadPoolExecutor
from typing import List, Optional, Union, Dict, Any
import re

import httpx
import rich.progress
from retrying import retry  # type: ignore

from capella_console_client.logconf import logger
from capella_console_client.hooks import (
    retry_if_httpx_status_error,
    log_attempt_delay,
)
from capella_console_client.exceptions import ConnectError


STAC_ID_REGEX = re.compile("^.*(CAPELLA_\\w+_\\w+_\\w+_\\d{14}_\\d{14}).*$")
PRODUCT_TYPE_REGEX = re.compile("^.*CAPELLA_\\w+_\\w+_(\\w+)_\\w+_\\d{14}_\\d{14}.*$")


@dataclass
class DownloadRequest:
    url: str
    local_path: Path
    asset_key: str
    stac_id: str = ""


progress_bar = rich.progress.Progress(
    rich.progress.TextColumn("[bold blue]{task.fields[filename]}", justify="left"),
    rich.progress.BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    rich.progress.DownloadColumn(),
    "•",
    rich.progress.TransferSpeedColumn(),
    "•",
    rich.progress.TimeRemainingColumn(),
)


def _flush_progress_bar(progress: rich.progress.Progress) -> None:
    for task_id in progress.task_ids:
        progress.remove_task(task_id)


def _gather_download_requests(
    assets_presigned: Dict[str, Any],
    local_dir: Union[Path, str] = Path(tempfile.gettempdir()),
    include: Union[List[str], str] = None,
    exclude: Union[List[str], str] = None,
    separate_dirs: bool = True,
) -> List[DownloadRequest]:
    local_dir = Path(local_dir)
    assert local_dir.exists(), f"{local_dir} does not exist"

    stac_id = _derive_stac_id(assets_presigned)

    if separate_dirs:
        local_dir /= stac_id
        local_dir.mkdir(exist_ok=True)

    logger.info(f"downloading product {stac_id} to {local_dir}")

    if include:
        include = _prep_include_exclude(include)
        logger.info(f"Only including assets {', '.join(include)}")

    if exclude:
        exclude = _prep_include_exclude(exclude)
        logger.info(f"Excluding assets {', '.join(exclude)}")

    # gather up paths
    download_requests = []
    for key, asset in assets_presigned.items():
        # white-listing asset
        if include and key not in include:
            continue

        # black-listing asset
        if exclude and key in exclude:
            continue

        local_path = local_dir / _get_filename(asset["href"])
        download_requests.append(
            DownloadRequest(
                stac_id=stac_id,
                asset_key=key,
                url=asset["href"],
                local_path=local_path,
            )
        )
    return download_requests


def _get_raster_href(assets_presigned: Dict[str, Any]) -> str:
    raster_asset = assets_presigned.get("HH")
    if raster_asset is None:
        raster_asset = assets_presigned["VV"]

    raster_asset_href: str = raster_asset["href"]
    return raster_asset_href


def _derive_stac_id(assets_presigned: Dict[str, Any]) -> str:
    raster_asset_href = _get_raster_href(assets_presigned)
    try:
        stac_id: str = STAC_ID_REGEX.findall(raster_asset_href)[0]
    except IndexError:
        raise ValueError(f"Could not derive STAC ID from {raster_asset_href}")
    return stac_id


def _derive_product_type(assets_presigned: Dict[str, Any]) -> str:
    raster_asset_href = _get_raster_href(assets_presigned)
    try:
        product_type: str = PRODUCT_TYPE_REGEX.findall(raster_asset_href)[0]
    except IndexError:
        raise ValueError(f"Could not derive product type from {raster_asset_href}")
    return product_type


def _filter_assets_by_product_types(
    assets_presigned: List[Dict[str, Any]], product_types: List[str]
) -> List[Dict[str, Any]]:
    logger.info(f'filtering by product_types: {", ".join(product_types)}')
    filtered_assets = []
    for cur_assets in assets_presigned:
        cur_product_type = _derive_product_type(cur_assets)
        if cur_product_type in product_types:
            filtered_assets.append(cur_assets)

    return filtered_assets


def _prep_include_exclude(filter_stmnt: Union[str, List[str]]) -> List[str]:
    if isinstance(filter_stmnt, str):
        filter_stmnt = [filter_stmnt]

    if "raster" in filter_stmnt:
        filter_stmnt.extend(["HH", "VV"])
        filter_stmnt.pop(filter_stmnt.index("raster"))

    return list(set(filter_stmnt))


def _perform_download(
    download_requests: List[DownloadRequest],
    override: bool,
    threaded: bool,
    show_progress: bool = False,
) -> Dict[str, Path]:

    local_paths_by_key = {}

    with progress_bar as progress:
        _flush_progress_bar(progress)

        # serially
        if not threaded:
            for dl_request in download_requests:
                local_paths_by_key[dl_request.asset_key] = _download_asset(
                    dl_request,
                    override=override,
                    show_progress=show_progress,
                    progress=progress,
                )

        # threaded
        else:
            with ThreadPoolExecutor(max_workers=len(download_requests)) as executor:
                # Start the load operations and mark each future with its URL
                futures_by_key = {}

                for dl_request in download_requests:
                    futures_by_key[dl_request.asset_key] = executor.submit(
                        _download_asset,
                        dl_request=dl_request,
                        override=override,
                        show_progress=show_progress,
                        progress=progress,
                    )

            for key, fut in futures_by_key.items():
                local_paths_by_key[key] = fut.result()

    return local_paths_by_key


def _download_asset(
    dl_request: DownloadRequest,
    override: bool,
    show_progress: bool,
    progress: rich.progress.Progress,
) -> Path:
    if dl_request.local_path is None:
        local_file = _get_filename(dl_request.url)
        dl_request.local_path = Path(tempfile.gettempdir()) / local_file

    dl_request.local_path = Path(dl_request.local_path)

    if not override and dl_request.local_path.exists():
        logger.info(f"already downloaded to {dl_request.local_path}")
        return dl_request.local_path

    try:
        asset_size = _get_asset_bytesize(dl_request.url)
    except Exception:
        asset_size = -1

    if not show_progress:
        size_suffix = f"({_sizeof_fmt(asset_size)})" if asset_size != -1 else ""
        logger.info(f"downloading to {dl_request.local_path} {size_suffix}")

    _fetch(dl_request, asset_size, show_progress, progress)

    if not show_progress:
        logger.info(f"successfully downloaded to {dl_request.local_path}")

    return dl_request.local_path


@retry(
    retry_on_exception=retry_if_httpx_status_error,
    wait_func=log_attempt_delay,
    wait_exponential_multiplier=2000,
    wait_exponential_max=16000,
)
def _fetch(
    dl_request: DownloadRequest,
    asset_size: int,
    show_progress: bool,
    progress: rich.progress.Progress,
):
    try:
        with open(dl_request.local_path, "wb") as f:
            with httpx.stream("GET", dl_request.url) as response:
                response.raise_for_status()
                if show_progress:
                    download_task_id = _register_progress_task(dl_request, progress, asset_size)

                for chunk in response.iter_bytes():
                    f.write(chunk)

                    if show_progress:
                        progress.update(download_task_id, completed=response.num_bytes_downloaded)
    except httpx.ConnectError as e:
        raise ConnectError(f"Could not connect to {dl_request.url}: {e}") from None

    return dl_request.local_path


def _register_progress_task(
    dl_request: DownloadRequest, progress: rich.progress.Progress, asset_size: int
) -> rich.progress.TaskID:
    file_name_str = str(dl_request.local_path)
    if dl_request.local_path and isinstance(dl_request.local_path, Path):
        file_name_str = dl_request.local_path.name
    download_task_id = progress.add_task("Download", total=asset_size, filename=file_name_str)
    return download_task_id


def _get_filename(pre_signed_url: str) -> str:
    return Path(urlparse(pre_signed_url).path).name


def _get_asset_bytesize(pre_signed_url: str) -> int:
    """get size in bytes of `pre_signed_url`"""
    try:
        with httpx.stream("GET", pre_signed_url) as resp:
            total_size = int(resp.headers["Content-Length"])
    except httpx.ConnectError as e:
        raise ConnectError(f"Could not connect to {pre_signed_url}: {e}") from None
    return total_size


def _sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f} {unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f} Yi{suffix}"



from typing import List, Dict, Any
from uuid import UUID
import typer
import questionary

from capella_console_client.cli.config import (
    CURRENT_SETTINGS,
)
from capella_console_client.cli.client_singleton import CLIENT
from capella_console_client.enumerations import BaseEnum
from capella_console_client.cli.visualize import (
    show_orders_tabulated,
    show_order_review_tabulated,
)
from capella_console_client.cli.validate import _no_selection_bye
from capella_console_client.cli.user_searches.core import _load_and_prompt, SearchEntity
from capella_console_client.cli.info import download_hint

app = typer.Typer(help="explore order history")


class PostOrderListActions(str, BaseEnum):
    reorder = "reorder"
    quit = "quit"

    @classmethod
    def _select_order(cls, question, orders):
        order_id_choices = [order["orderId"] for order in orders]
        selected_order_id = questionary.autocomplete(
            question,
            choices=order_id_choices,
            validate=lambda x: x in order_id_choices,
        ).ask()
        _no_selection_bye(selected_order_id)

        idx = order_id_choices.index(selected_order_id)
        return orders[idx]

    @classmethod
    def prompt_and_reorder(cls, orders):
        selected_order = cls._select_order(question="Specify the orderId you'd like to resubmit:", orders=orders)

        stac_ids = [item["granuleId"] for item in selected_order["items"]]
        order_id = CLIENT.submit_order(stac_ids=stac_ids)
        download_hint(order_id)
        return order_id


def _prompt_post_order_list_actions(orders: List[Dict[str, Any]]):
    choices = list(PostOrderListActions)

    action_selection = None
    while action_selection != PostOrderListActions.quit:
        action_selection = questionary.select(
            "Anything you'd like to do now?",
            choices=choices,
        ).ask()
        _no_selection_bye(action_selection)

        if action_selection == PostOrderListActions.reorder:
            PostOrderListActions.prompt_and_reorder(orders)


def _list_orders_and_tabulate(
    is_active: bool, limit: int = CURRENT_SETTINGS["order_list_limit"]
) -> List[Dict[str, Any]]:
    orders = CLIENT.list_orders(is_active=is_active)
    if not orders:
        typer.echo("Currently no orders available")
        raise typer.Exit(0)

    orders = sorted(orders, key=lambda x: x["orderDate"], reverse=True)[:limit]
    show_orders_tabulated(orders)
    return orders


@app.command("list")
def list_orders(
    is_active: bool = typer.Option(False, "--active", help="only show active (non-expired) orders"),
    limit: int = typer.Option(
        CURRENT_SETTINGS["order_list_limit"],
        help="limit orders to display (up to 300 currently)",
    ),
):
    """
    list your orders by orderDate
    """
    orders = _list_orders_and_tabulate(is_active, limit)
    _prompt_post_order_list_actions(orders)


@app.command("review")
def review(
    is_active: bool = typer.Option(False, "--active", help="only show active (non-expired) orders"),
    from_saved: bool = typer.Option(False, "--from-saved", help="from previously saved search result"),
):
    """
    review orders
    """
    if from_saved:
        my_searches, selection = _load_and_prompt(
            "Which saved search result would you like to use?",
            search_entity=SearchEntity.result,
            multiple=False,
        )
        stac_ids = my_searches[selection]["data"]  # type: ignore
    else:
        orders = _list_orders_and_tabulate(is_active)
        selected_order = PostOrderListActions._select_order(
            question="Which order would you like to review:", orders=orders
        )
        stac_ids = [item["granuleId"] for item in selected_order["items"]]

    order_review = CLIENT.review_order(stac_ids=stac_ids)
    show_order_review_tabulated(order_review)


@app.command("reorder")
def reorder(order_id: UUID):
    """
    re-order by order ID
    """
    order = CLIENT.list_orders(str(order_id))[0]
    stac_ids = [item["granuleId"] for item in order["items"]]
    new_order_id = CLIENT.submit_order(stac_ids=stac_ids)
    download_hint(new_order_id)



"""
CLI / interactive (questionary) specific validation
"""

import json
from pathlib import Path
import re
from dateutil.parser import parse, ParserError
from typing import Union

import typer

from capella_console_client.validate import _validate_uuid as _validate_core_uuid
from capella_console_client.assets import STAC_ID_REGEX


EMAIL_REGEX = re.compile(r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b")


def _must_be_type(type):
    def _must_be_type_impl(val):
        try:
            type(val.strip())
        except Exception:
            return f"please provide a valid {type}"
        return True

    return _must_be_type_impl


def _validate_bbox(val):
    err_msg = "please specify as bbox list [<lon upper left>, <lat upper left>, <lon lower right>, <lat lower right>], e.g. [-122.4, 46.9, -124.9, 48.5]"
    if not val:
        return err_msg
    try:
        parsed = json.loads(val)
    except:
        return err_msg

    if not isinstance(parsed, list):
        return err_msg

    if not len(parsed) == 4:
        return err_msg
    return True


def _validate_uuid(val):
    err_msg = "please specify a valid uuid (e.g. aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee)"
    try:
        _validate_core_uuid(val)
    except ValueError as e:
        return err_msg

    return True


def _validate_dir_exists(path_str: str) -> Union[bool, str]:
    err_msg = "please specify an existing directory path"
    p = Path(path_str).resolve()
    if p.exists() and p.is_dir():
        return True
    return err_msg


def _validate_out_path(path_str: str) -> bool:
    p = Path(path_str).resolve()
    return p.is_absolute() and p.parent.exists() and not p.is_dir()


def _validate_email(email: str):
    err_msg = "please specify a valid email"
    if EMAIL_REGEX.fullmatch(email):
        return True
    return err_msg


def _validate_datetime(dt_str: str):
    err_msg = "please specify a valid UTC date(time) (e.g. 2020-08-14 or 2020-08-14 12:00:00)"
    if not dt_str or len(dt_str) < 10:
        return err_msg

    try:
        parse(dt_str)
    except ParserError:
        return err_msg
    return True


def _parse_str_collection(list_str):
    _cur = list_str.strip()
    if not _cur.startswith("[") or not _cur.endswith("]"):
        # raw parse
        separator = "," if "," in _cur else " "
        stac_ids = _cur.split(separator)
    else:
        stac_ids = json.loads(_cur.replace("'", '"'))
    return stac_ids


def _validate_stac_ids(stac_id_str: str):
    err_msg = "please specify one or more , or whitespace separated STAC Ids (e.g. CAPELLA_C03_SM_GEO_HH_20210512034455_20210512034459,CAPELLA_C03_SP_GEO_HH_20210511101416_20210511101439)"

    stac_ids = _parse_str_collection(stac_id_str)
    if all(STAC_ID_REGEX.match(stac_id) for stac_id in stac_ids):
        return True

    return err_msg


def _validate_collections(stac_id_str: str):
    err_msg = "please specify one or more , or whitespace separated collections (e.g. capella-open-data)"

    stac_ids = _parse_str_collection(stac_id_str)
    if all(STAC_ID_REGEX.match(stac_id) for stac_id in stac_ids):
        return True

    return err_msg


def get_validator(field: str):
    custom_validator = {
        "bbox": _validate_bbox,
        "collect_id": _validate_uuid,
        "datetime": _validate_datetime,
        "ids": _validate_stac_ids,
    }.get(field)

    if custom_validator:
        return custom_validator

    type_caster = get_caster(field)

    if type_caster:
        return _must_be_type(type_caster)

    return None


def _cast_dt(val):
    return parse(val).strftime("%Y-%m-%dT%H:%M:%SZ")


def get_caster(field: str):
    return {
        "bbox": json.loads,
        "billable_area": int,
        "center_frequency": float,
        "incidence_angle": float,
        "limit": int,
        "look_angle": float,
        "looks_azimuth": int,
        "looks_equivalent_number": int,
        "looks_range": int,
        "pixel_spacing_azimuth": float,
        "pixel_spacing_range": float,
        "polarizations": float,
        "resolution_azimuth": float,
        "resolution_ground_range": float,
        "resolution_range": float,
        "squint_angle": float,
        "datetime": _cast_dt,
        "ids": _parse_str_collection,
        "collections": _parse_str_collection,
    }.get(field)


def _no_selection_bye(selection, info_msg=None):
    if not info_msg:
        info_msg = "nothing selected ... bye"

    if selection in (None, []):
        typer.echo(info_msg)
        raise typer.Exit(code=1)


def _at_least_one_selected(val):
    err_msg = "Please select at least one option"
    if len(val) < 1:
        return err_msg
    return True



import typer

import capella_console_client.cli.user_searches.my_search_queries
import capella_console_client.cli.user_searches.my_search_results

app = typer.Typer(help="manage my-searches (queries and results)")
app.add_typer(capella_console_client.cli.user_searches.my_search_queries.app, name="queries")
app.add_typer(capella_console_client.cli.user_searches.my_search_results.app, name="results")



import typer


def no_data_info(search_entity):
    typer.secho(f"No search {search_entity.name} currently saved!\n", bold=True)
    typer.echo(f"Issue")
    typer.secho(f"\tcapella-console-wizard workflows search", bold=True)
    typer.echo(f"\nin order to search and save your first search {search_entity.name}.")
    raise typer.Exit(code=1)


def download_hint(order_id: str):
    typer.echo(f"Issue")
    typer.secho("\tcapella-console-wizard checkout", bold=True)
    typer.echo(
        f"\n select order 'select existing order' and provide order id {order_id} in order to download all products of the order."
    )


def my_search_entity_info(identifier: str):
    txt = f"""Added '{identifier}' to my-search-results and my-search-queries"
Issue

\tcapella-console-wizard my-search-results list

    or

\tcapella-console-wizard my-search-queries list

in order to list your saved search results or queries.
"""
    typer.echo(txt)



CONSOLE_API_URL = "https://api.capellaspace.com"
DEFAULT_TIMEOUT = 60
DEFAULT_PAGE_SIZE = 1000
DEFAULT_MAX_FEATURE_COUNT = 500


SUPPORTED_SEARCH_FIELDS = {
    "bbox",
    "intersects",
    "collections",
    "ids",
    "limit",
}


SUPPORTED_QUERY_FIELDS = {
    "center_frequency",
    "collect_id",
    "constellation",
    "datetime",
    "frequency_band",
    "incidence_angle",
    "instruments",
    "instrument_mode",
    "look_angle",
    "looks_azimuth",
    "looks_equivalent_number",
    "looks_range",
    "observation_direction",
    "orbit_state",
    "pixel_spacing_azimuth",
    "pixel_spacing_range",
    "platform",
    "product_category",
    "product_type",
    "polarizations",
    "resolution_azimuth",
    "resolution_ground_range",
    "resolution_range",
    "squint_angle",
    "orbital_plane",
    "billable_area",
}

ALL_SUPPORTED_FIELDS = SUPPORTED_SEARCH_FIELDS | SUPPORTED_QUERY_FIELDS

ALL_SUPPORTED_SORTBY = ALL_SUPPORTED_FIELDS | {"id"}

OPERATOR_SUFFIXES = {
    "eq",
    "in",
    "gt",
    "gte",
    "lt",
    "lte",
}  # , "startsWith", "endsWith"} not supported by current STAC server


STAC_PREFIXED_BY_QUERY_FIELDS = {
    "collect_id": "capella:collect_id",
    "product_category": "capella:product_category",
    "resolution_ground_range": "capella:resolution_ground_range",
    "squint_angle": "capella:squint_angle",
    "constellation": "constellation",
    "instruments": "instruments",
    "platform": "platform",
    "center_frequency": "sar:center_frequency",
    "frequency_band": "sar:frequency_band",
    "instrument_mode": "sar:instrument_mode",
    "looks_azimuth": "sar:looks_azimuth",
    "looks_equivalent_number": "sar:looks_equivalent_number",
    "looks_range": "sar:looks_range",
    "observation_direction": "sar:observation_direction",
    "pixel_spacing_azimuth": "sar:pixel_spacing_azimuth",
    "pixel_spacing_range": "sar:pixel_spacing_range",
    "polarizations": "sar:polarizations",
    "product_type": "sar:product_type",
    "resolution_azimuth": "sar:resolution_azimuth",
    "resolution_range": "sar:resolution_range",
    "orbit_state": "sat:orbit_state",
    "incidence_angle": "view:incidence_angle",
    "look_angle": "view:look_angle",
    "orbital_plane": "capella:orbital_plane",
    "billable_area": "capella:billable_area",
}



from typing import Tuple, Dict, Any, List
from datetime import datetime

import typer
import questionary

from capella_console_client.enumerations import BaseEnum
from capella_console_client.cli.validate import _no_selection_bye
from capella_console_client.cli.cache import CLICache
from capella_console_client.cli.info import no_data_info


class SearchEntity(str, BaseEnum):
    result = 1
    query = 2


def _get_load_fct(search_entity: SearchEntity):
    return {
        SearchEntity.result: CLICache.load_my_search_results,
        SearchEntity.query: CLICache.load_my_search_queries,
    }[search_entity]


def _get_write_fct(search_entity):
    return {
        SearchEntity.result: CLICache.write_my_search_results,
        SearchEntity.query: CLICache.write_my_search_queries,
    }[search_entity]


def _load_and_prompt(
    question: str,
    search_entity: SearchEntity,
    multiple: bool = True,
) -> Tuple[Dict[str, Any], List[str]]:

    saved = _get_load_fct(search_entity)()

    if not saved:
        no_data_info(search_entity)

    typer.echo("\n\n")

    question_cls = questionary.checkbox if multiple else questionary.select
    selection = question_cls(message=question, choices=saved).ask()  # type: ignore
    _no_selection_bye(selection)
    return (saved, selection)


def rename_search_entity(search_entity: SearchEntity):
    saved, selection = _load_and_prompt(f"Which saved {search_entity.name} would you like to rename?", search_entity)

    change_cnt = 0
    for selected in selection:
        new_name = questionary.text(
            message=f"Rename {selected} to:",
            default=selected,
            validate=lambda x: x is not None and len(x) > 0,
        ).ask()

        if new_name != selected:
            val = saved[selected]
            saved[new_name] = val
            saved[new_name]["updated_at"] = str(datetime.now())[:-7]
            del saved[selected]
            change_cnt += 1

    if change_cnt > 0:
        _get_write_fct(search_entity)(saved)
        typer.echo(f"Renamed {change_cnt} search {search_entity.name}")



from typing import Dict, List, Any, Optional

from capella_console_client.logconf import logger


def _sort_stac_items(items: List[Dict[str, Any]], stac_ids: List[str]) -> List[Dict[str, Any]]:
    """
    sort items by stac_ids

    Args:
        items (List[Dict[str, Any]]): stac items
        stac_ids (List[str]): stac ids to sort by

    Returns:
        List[Dict[str, Any]]: stac items sorted by stac_ids
    """
    if len(stac_ids) != len(items):
        logger.warning(f"wrong size stac_ids ({len(stac_ids)} instead of {len(items)})... omitting sort ")
        return items

    sorted = []
    item_ids = [s["id"] for s in items]
    for stac_id in stac_ids:
        try:
            cur = items[item_ids.index(stac_id)]
        except ValueError:
            continue
        sorted.append(cur)

    not_in_items = [item for item in items if item["id"] not in stac_ids]
    sorted.extend(not_in_items)
    return sorted



from typing import List
import questionary


def get_first_checked(choices: List[questionary.Choice], prev_search=None) -> questionary.Choice:
    first_checked = choices[0]
    if prev_search:
        first_checked = next(c for c in choices if c.checked)
    return first_checked



import os
from typing import List, Dict, Any, Optional, Tuple
import json
from collections import defaultdict

import typer
import questionary

from capella_console_client.enumerations import BaseEnum
from capella_console_client.search import SearchResult
from capella_console_client.cli.client_singleton import CLIENT
from capella_console_client.cli.validate import (
    get_validator,
    get_caster,
    _validate_out_path,
    _no_selection_bye,
    _at_least_one_selected,
)
from capella_console_client.cli.cache import CLICache
from capella_console_client.cli.config import (
    CLI_SUPPORTED_SEARCH_FILTERS,
    CURRENT_SETTINGS,
    PROMPT_OPERATORS,
    ENUM_CHOICES_BY_FIELD_NAME,
)
from capella_console_client.cli.user_searches.my_search_results import _load_and_prompt
from capella_console_client.cli.user_searches.core import SearchEntity
from capella_console_client.cli.visualize import show_tabulated
from capella_console_client.cli.settings import _prompt_search_result_headers
from capella_console_client.cli.info import my_search_entity_info
from capella_console_client.cli.prompt_helpers import get_first_checked


# TODO: autocomplete option
def interactive_search():
    search_query = _prompt_search_filters()
    choices = list(PostSearchActions)
    del choices[choices.index(PostSearchActions.continue_flow)]
    search_and_post_actions(search_query, choices=choices)


class STACQueryPayload(dict):
    OPS_MAP = {
        "=": "eq",
        ">": "gt",
        ">=": "gte",
        "<": "lt",
        "<=": "lte",
        "in": "in",
    }

    REV_OPS_MAP = {
        "eq": "=",
        "gt": ">",
        "gte": ">=",
        "lt": "<",
        "lte": "<=",
        "in": "in",
    }

    def __str__(self):
        _filters = []
        for k, v in self.items():
            if isinstance(v, list):
                cur = f"{k}{'|'.join(map(str, v))}"
            else:
                cur = f"{k}{v}"
            _filters.append(cur)

        return "-".join(_filters)

    @classmethod
    def unflatten(cls, other) -> "STACQueryPayload":
        _con = defaultdict(list)
        for k, v in other.items():
            if "__" in k:
                key, op = k.split("__")
                op = STACQueryPayload.REV_OPS_MAP[op]
                _con[key].append((op, v))
            else:
                _con[k].append(("=", v))
        return cls(_con)

    def flatten(self, other):
        _con = {}
        for field, op_val_tuples in other.items():
            for op, val in op_val_tuples:
                _con.add(field, op, val)
        return _con

    def add(self, field: str, search_op: str, value: Any):
        if not search_op or search_op == "=":
            field_desc = field
        else:
            field_desc = f"{field}__{self.OPS_MAP[search_op]}"
        self[field_desc] = value


def _prompt_search_operator(field: str, prev_selection: List[str]) -> List[str]:
    choices = [questionary.Choice(cur, checked=cur in prev_selection) for cur in STACQueryPayload.OPS_MAP.keys()]
    operators = questionary.checkbox(
        f"{field}:",
        choices=choices,
        initial_choice=get_first_checked(choices, prev_selection),
    ).ask()
    _no_selection_bye(operators, "Please select at least one search operator")
    return operators


def _prompt_enum_choices(field: str, init: Any = None) -> Optional[Dict[str, Any]]:
    if init is None:
        init = []

    if field not in ENUM_CHOICES_BY_FIELD_NAME:
        return None

    choices = [  # type: ignore
        questionary.Choice(e.value, checked=e.value in init) for e in ENUM_CHOICES_BY_FIELD_NAME[field]
    ]

    choices = questionary.checkbox(
        f"{field}:", choices=choices, initial_choice=get_first_checked(choices, init)
    ).ask()  # type: ignore
    _no_selection_bye(choices)
    return {field: choices}


def _prompt_operator_value(field: str, search_op: str, init: Any = ""):
    suffix = f"[{search_op}]" if search_op else ""
    str_val = questionary.text(
        message=f"{field} {suffix}:",
        default=str(init),
        validate=get_validator(field),
    ).ask()
    _no_selection_bye(str_val)
    cast_fct = get_caster(field)
    if cast_fct:
        return cast_fct(str_val)
    return str_val


def _prompt_search_filters(prev_search: STACQueryPayload = None) -> STACQueryPayload:
    if prev_search is None:
        prev_search = STACQueryPayload()

    choices = [questionary.Choice(cur, checked=cur in prev_search) for cur in CLI_SUPPORTED_SEARCH_FILTERS]

    search_filter_names = questionary.checkbox(
        "Select your search filters:",
        choices=choices,
        initial_choice=get_first_checked(choices, prev_search),
        validate=_at_least_one_selected,
    ).ask()
    _no_selection_bye(search_filter_names, "Please select at least one search filter")

    query = STACQueryPayload()
    for field in search_filter_names:
        prev_selected_ops = {x[0]: x[1] for x in prev_search.get(field, [])}
        cur = _prompt_enum_choices(field, init=prev_selected_ops.get("="))

        # enum takes precedence
        if cur:
            query.update(cur)
            continue

        if field in PROMPT_OPERATORS:
            operators = _prompt_search_operator(field, list(prev_selected_ops.keys()))
        else:
            operators = ["="]

        for search_op in operators:
            init = prev_selected_ops.get(search_op, "")
            value = _prompt_operator_value(field, search_op, init)
            query.add(field, search_op, value)

    if "limit" not in query:
        query["limit"] = CURRENT_SETTINGS["limit"]

    if "constellation" not in query:
        query["constellation"] = "capella"

    return query


class PostSearchActions(str, BaseEnum):
    refine_search = "refine search"
    adjust_headers = "change result table headers"
    save_current_search = "save search query and result into my-search-results | my-search-queries"
    export_json = "export STAC items of search as .json"
    continue_flow = "continue"
    quit = "quit"

    @classmethod
    def save_search(cls, result: SearchResult, search_kwargs: STACQueryPayload):
        identifier = questionary.text(
            message="Please provide an identifier for your search:",
            default=str(search_kwargs),
            validate=lambda x: x is not None and len(x) > 0,
        ).ask()
        CLICache.update_my_search_results(identifier, result.stac_ids, is_new=True)
        CLICache.update_my_search_queries(identifier, search_kwargs, is_new=True)  # type: ignore
        my_search_entity_info(identifier)

    @classmethod
    def export_search(cls, result: SearchResult, search_kwargs: STACQueryPayload) -> str:
        default = CURRENT_SETTINGS["out_path"]
        if default[-1] != os.sep:
            default += os.sep
        default += f"{search_kwargs}.json"

        path = questionary.path(
            message="Please provide path and filename where you want to save the stac items .json",
            default=default,
            validate=_validate_out_path,
        ).ask()
        _no_selection_bye(path, "Please provide a path")

        with open(path, "w") as fp:
            json.dump(result, fp)
        typer.echo(f"Saved {len(result)} STAC items to {path}")
        return path

    @classmethod
    def refine_search_cmd(cls, prev_search: STACQueryPayload) -> Tuple[STACQueryPayload, SearchResult]:
        prev_search.pop("constellation", None)
        if prev_search["limit"][0][1] == CURRENT_SETTINGS["limit"]:
            prev_search.pop("limit")

        typer.echo(f"Refining\n\t{json.dumps(prev_search)}")
        search_query = _prompt_search_filters(prev_search=prev_search)
        stac_items = CLIENT.search(**search_query)
        return (search_query, stac_items)

    @classmethod
    def _get_choices(cls, results_found: bool) -> List["PostSearchActions"]:
        if results_found:
            return list(PostSearchActions)
        else:
            return [PostSearchActions.refine_search, PostSearchActions.quit]


def search_and_post_actions(search_query: STACQueryPayload, choices: List[PostSearchActions] = None):
    result = CLIENT.search(**search_query)
    if result:
        show_tabulated(result, show_row_number=True)

    result = _prompt_post_search_actions(result, search_query, choices)
    return result


def _prompt_post_search_actions(
    result: SearchResult,
    search_kwargs: STACQueryPayload,
    choices: List[PostSearchActions] = None,
):
    if not choices:
        choices = PostSearchActions._get_choices(results_found=len(result) > 0)
    action_selection = None

    while action_selection != PostSearchActions.quit:
        action_selection = questionary.select(
            "Anything you'd like to do now?",
            choices=choices,
        ).ask()
        _no_selection_bye(action_selection)

        if action_selection == PostSearchActions.adjust_headers:
            search_headers = _prompt_search_result_headers()
            CURRENT_SETTINGS["search_headers"] = search_headers
            CLICache.write_user_settings("search_headers", search_headers)
            show_tabulated(result, search_headers, show_row_number=True)

        if action_selection == PostSearchActions.save_current_search:
            PostSearchActions.save_search(result, search_kwargs)
            del choices[choices.index(PostSearchActions.save_current_search)]

        if action_selection == PostSearchActions.export_json:
            path = PostSearchActions.export_search(result, search_kwargs)
            if questionary.confirm("Would you like to open it?").ask():
                os.system(f"open {path}")

        if action_selection == PostSearchActions.refine_search:
            prev_search = STACQueryPayload.unflatten(search_kwargs)
            search_kwargs, result = PostSearchActions.refine_search_cmd(prev_search)
            show_tabulated(result, show_row_number=True)
            choices = PostSearchActions._get_choices(results_found=bool(result))

        if action_selection == PostSearchActions.continue_flow:
            return result


def from_saved():
    """
    select and show STAC items of saved search
    """
    entity = questionary.select(
        "Would you like to use a saved query or a saved result",
        choices=[SearchEntity.query.name, SearchEntity.result.name],
    ).ask()
    _no_selection_bye(entity)

    search_entity = SearchEntity[entity]
    saved, selection = _load_and_prompt(
        f"Which saved {entity} would you like to use?",
        search_entity=search_entity,
        multiple=False,
    )

    if search_entity == SearchEntity.result:
        search_query = dict(ids=saved[selection]["data"])
    else:
        search_query = saved[selection]["data"]

    search_and_post_actions(search_query)



"""
try to copy s3://cirrus-prod-data-27qi4kja7o84/static-icons/ACD.png to govcloud bucket    
"""

import boto3

if __name__ == "__main__":

    GOVCLOUD_BUCKET = "<insert>"
    GOVCLOUD_PREFIX = "<insert-no-trailing-/>"
    GOVCLOUD_AWS_CLI_PROFILE_NAME = "<insert>"

    CIRRUS_PROD_DATA_BUCKET = "cirrus-prod-data-27qi4kja7o84"
    CIRRUS_ASSET_S3_KEY = "static-icons/ACD.png"

    source_s3 = boto3.Session(profile_name="e84-prod").client("s3")

    # check read access
    head_ret = source_s3.head_object(Bucket=CIRRUS_PROD_DATA_BUCKET, Key=CIRRUS_ASSET_S3_KEY)
    print(head_ret["ResponseMetadata"])

    dest_s3 = boto3.Session(profile_name=GOVCLOUD_AWS_CLI_PROFILE_NAME).client("s3")
    dest_s3_key = f"{GOVCLOUD_PREFIX}/{CIRRUS_ASSET_S3_KEY}"

    # pretty certain this will fail
    dest_s3.copy(
        CopySource={"Bucket": CIRRUS_PROD_DATA_BUCKET, "Key": CIRRUS_ASSET_S3_KEY},
        Bucket=GOVCLOUD_BUCKET,
        Key=dest_s3_key,
        SourceClient=source_s3,
    )

    # check if exists
    gc_head_ret = source_s3.head_object(Bucket=GOVCLOUD_BUCKET, Key=dest_s3_key)
    print(gc_head_ret["ResponseMetadata"])



import logging
import sys

from datetime import datetime
from typing import List, Dict, Any, Union, Optional, no_type_check, Tuple
from collections import defaultdict
from pathlib import Path
import tempfile

import dateutil.parser

from capella_console_client.config import CONSOLE_API_URL
from capella_console_client.session import CapellaConsoleSession
from capella_console_client.logconf import logger
from capella_console_client.exceptions import (
    InsufficientFundsError,
    OrderRejectedError,
    NoValidStacIdsError,
    TaskNotCompleteError,
)

from capella_console_client.assets import (
    _perform_download,
    DownloadRequest,
    _gather_download_requests,
    _get_asset_bytesize,
    _derive_stac_id,
    _filter_assets_by_product_types,
)
from capella_console_client.search import StacSearch, SearchResult
from capella_console_client.validate import (
    _validate_uuid,
    _validate_stac_id_or_stac_items,
    _validate_and_filter_product_types,
    _validate_and_filter_asset_types,
    _validate_and_filter_stac_ids,
)
from capella_console_client.sort import _sort_stac_items


class CapellaConsoleClient:
    """
    API client for https://api.capellaspace.com.

    API docs: https://docs.capellaspace.com/accessing-data/searching-for-data

    Args:
        email: email on api.capellaspace.com
        password: password on api.capellaspace.com
        token: valid JWT access token
        verbose: flag to enable verbose logging
        no_token_check: does not check if provided JWT token is valid
        base_url: Capella console API base URL override
        search_url: Capella catalog/search/ override
        no_auth: bypass authentication

    NOTE:
        not providing either email and password or a jwt token for authentication
        will prompt you for email and password, which is not what you want in a script

    NOTE: Precedence order (high to low)
        1. email and password
        2. JWT token
    """

    def __init__(
        self,
        email: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        verbose: bool = False,
        no_token_check: bool = False,
        base_url: Optional[str] = CONSOLE_API_URL,
        search_url: Optional[str] = None,
        no_auth: bool = False,
    ):
        self._set_verbosity(verbose)
        self._sesh = CapellaConsoleSession(base_url=base_url, search_url=search_url, verbose=verbose)

        if not no_auth:
            self._sesh.authenticate(email, password, token, no_token_check)

    def _set_verbosity(self, verbose: bool = False):
        self.verbose = verbose
        logger.setLevel(logging.WARNING)
        if verbose:
            logger.setLevel(logging.INFO)

    # USER
    def whoami(self) -> Dict[str, Any]:
        """
        display user info

        Returns:
            Dict[str, Any]: return of GET /user
        """
        resp = self._sesh.get("/user")
        return resp.json()

    # TASKING
    def list_tasking_requests(
        self,
        *tasking_request_ids: Optional[str],
        for_org: Optional[bool] = False,
        status: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        list tasking requests

        Args:
            tasking_request_ids: list only specific tasking_request_ids (variadic, specify multiple)
            for_org: list all tasking requests of your organization (instead of only yours) - **requires** organization index/ admin permission
            status: list only tasking requests that are in this/ have passed this status, e.g. completed

        Returns:
            List[Dict[str, Any]]: metadata of tasking requests
        """
        tasking_request_meta = []

        if tasking_request_ids:
            for t_req_id in tasking_request_ids:
                _validate_uuid(t_req_id)

        # get selected
        if tasking_request_ids:
            tasking_request_meta = [self.get_task(t_req_id) for t_req_id in tasking_request_ids]  # type: ignore
            if status:
                tasking_request_meta = [
                    t_meta for t_meta in tasking_request_meta if self._task_contains_status(t_meta, status)
                ]
            return tasking_request_meta

        if for_org:
            params = {"organizationId": self._sesh.organization_id}
        else:
            params = {"customerId": self._sesh.customer_id}

        resp = self._sesh.get("/tasks", params=params)
        tasking_request_meta = resp.json()

        if status:
            tasking_request_meta = [
                t_meta for t_meta in tasking_request_meta if self._task_contains_status(t_meta, status)
            ]
        return tasking_request_meta

    def get_task(self, tasking_request_id: str) -> Dict[str, Any]:
        """
        fetch task for the specified `tasking_request_id`

        Args:
            tasking_request_id: tasking request UUID

        Returns:
            Dict[str, Any]: task metadata
        """
        task_response = self._sesh.get(f"/task/{tasking_request_id}")
        return task_response.json()

    def _task_contains_status(self, task: Dict[str, Any], status_name: str) -> bool:
        return status_name.lower() in (s["code"] for s in task["properties"]["statusHistory"])

    def is_task_completed(self, task: Dict[str, Any]) -> bool:
        """
        check if a task has completed
        """
        return self._task_contains_status(task, "completed")

    def get_collects_for_task(self, tasking_request_id: str) -> List[Dict[str, Any]]:
        """
        get all the collects associated with this task (see :py:meth:`get_task()`)

        Args:
            task: task metadata - return of :py:meth:`get_task()`

        Returns:
            List[Dict[str, Any]]: collect metadata associated
        """
        task = self.get_task(tasking_request_id)
        tasking_request_id = task["properties"]["taskingrequestId"]
        if not self.is_task_completed(task):
            raise TaskNotCompleteError(f"TaskingRequest<{tasking_request_id}> is not in completed state")

        collects_list_resp = self._sesh.get(f"/collects/list/{tasking_request_id}")

        return collects_list_resp.json()

    # ORDER
    def list_orders(self, *order_ids: Optional[str], is_active: Optional[bool] = False) -> List[Dict[str, Any]]:
        """
        list orders

        Args:
            order_id: list only specific orders (variadic, specify multiple)
            is_active: list only active (non-expired) orders

        Returns:
            List[Dict[str, Any]]: metadata of orders
        """
        orders = []

        if order_ids:
            for order_id in order_ids:
                _validate_uuid(order_id)

        # prefilter non expired
        if is_active:
            orders = _get_non_expired_orders(session=self._sesh)
            if order_ids:
                set_order_ids = set(order_ids)
                orders = [o for o in orders if o["orderId"] in set_order_ids]
        else:
            # list all orders
            if not order_ids:
                params = {
                    "customerId": self._sesh.customer_id,
                }
                resp = self._sesh.get("/orders", params=params)
                orders = resp.json()

            # list specific orders
            else:
                for order_id in order_ids:
                    resp = self._sesh.get(f"/orders/{order_id}")
                    orders.append(resp.json())

        return orders

    def get_stac_items_of_order(self, order_id: str, ids_only: bool = False) -> Union[List[str], SearchResult]:
        """
        get stac items of an existing order

        Args:
            order_id: order id
        """
        _validate_uuid(order_id)
        order_meta = self.list_orders(order_id)[0]

        stac_ids = [item["granuleId"] for item in order_meta["items"]]
        if ids_only:
            return stac_ids

        return self.search(ids=stac_ids)

    def review_order(
        self,
        stac_ids: Optional[List[str]] = None,
        items: Optional[Union[List[Dict[str, Any]], SearchResult]] = None,
    ) -> Dict[str, Any]:
        stac_ids = _validate_stac_id_or_stac_items(stac_ids, items)

        logger.info(f"reviewing order for {', '.join(stac_ids)}")

        stac_items = items  # type: ignore
        if not items:
            stac_items = self.search(ids=stac_ids)

        if not stac_items:
            raise NoValidStacIdsError(f"No valid STAC IDs in {', '.join(stac_ids)}")

        # review order
        order_payload = self._construct_order_payload(stac_items)
        review_order_response = self._sesh.post("/orders/review", json=order_payload).json()

        if not review_order_response.get("authorized", False):
            raise InsufficientFundsError(review_order_response["authorizationDenialReason"]["message"])
        return review_order_response

    def submit_order(
        self,
        stac_ids: Optional[List[str]] = None,
        items: Optional[Union[List[Dict[str, Any]], SearchResult]] = None,
        check_active_orders: bool = False,
        omit_search: bool = False,
        omit_review: bool = False,
    ) -> str:
        """
        submit an order by `stac_ids` or `items`.

        NOTE: Precedence order (high to low)
            1. stac_ids
            2. items

        Args:
            stac_ids: STAC IDs that active order should include
            items: STAC items, returned by :py:meth:`search`
            check_active_orders: check if any active order containing ALL `stac_ids` is available
                if True: returns that order ID
                if False: submits a new order and returns new order ID
            omit_search: omit search to ensure provided STAC IDs are valid - only works if `items` are provided
            omit_review: omit review stage
            Returns:
                str: order UUID
        """
        stac_ids = _validate_stac_id_or_stac_items(stac_ids, items)

        if check_active_orders:
            order_id = self._find_active_order(stac_ids)
            if order_id is not None:
                logger.info(f"found active order {order_id}")
                return order_id

        if stac_ids and not omit_search:
            stac_items = self.search(ids=stac_ids)
        else:
            if omit_search and not items:
                logger.warning("setting omit_search=True only works in combination providing items instead of stac_ids")
                stac_items = self.search(ids=stac_ids)
            else:
                stac_items = items  # type: ignore

        if not stac_items:
            raise NoValidStacIdsError(f"No valid STAC IDs in {', '.join(stac_ids)}")

        if not omit_review:
            self.review_order(items=stac_items)

        logger.info(f"submitting order for {', '.join(stac_ids)}")
        order_payload = self._construct_order_payload(stac_items)
        res_order = self._sesh.post("/orders", json=order_payload)

        con = res_order.json()
        order_id = con["orderId"]
        if con["orderStatus"] == "rejected":
            raise OrderRejectedError(f"Order for {', '.join(stac_ids)} rejected.")

        logger.info(f"successfully submitted order {order_id}")
        return order_id  # type: ignore

    def _construct_order_payload(self, stac_items):
        by_collect_id = defaultdict(list)
        for item in stac_items:
            by_collect_id[item["collection"]].append(item["id"])

        order_items = []
        for collection, stac_ids_of_coll in by_collect_id.items():
            order_items.extend([{"collectionId": collection, "granuleId": stac_id} for stac_id in stac_ids_of_coll])
        return {"items": order_items}

    def _find_active_order(self, stac_ids: List[str]) -> Optional[str]:
        """
        find active order containing ALL specified `stac_ids`

        Args:
            stac_ids: STAC IDs that active order should include
        """
        order_id = None
        active_orders = _get_non_expired_orders(session=self._sesh)
        if not active_orders:
            return None

        for ord in active_orders:
            granules = set([i["granuleId"] for i in ord["items"]])
            if granules.issuperset(stac_ids):
                order_id = ord["orderId"]
                break
        return order_id

    def get_presigned_assets(
        self,
        order_id: str,
        stac_ids: Optional[List[str]] = None,
        sort_by: Optional[List[str]] = None,
        assets_only: Optional[bool] = True,
    ) -> List[Dict[str, Any]]:
        """
        get presigned assets hrefs for all products contained in order

        Args:
            order_id: active order ID (see :py:meth:`submit_order`)
            stac_ids: filter presigned assets by STAC IDs
            sort_by: list of stac ids to sort by
            assets_only: return only list of STAC item assets

        Returns:
            List[Dict[str, Any]]: List of assets of respective product, e.g.

            .. highlight:: python
            .. code-block:: python

                [
                    {
                        "<asset_type>": {
                            "title": ...,
                            "href": ...,
                            "type": ...
                        },
                        ...
                    }
                ]

        """
        _validate_uuid(order_id)

        logger.info(f"getting presigned assets for order {order_id}")
        response = self._sesh.get(f"/orders/{order_id}/download")

        presigned_stac_items = response.json()

        # ensure sort
        sort_by = _validate_and_filter_stac_ids(sort_by)
        if sort_by:
            presigned_stac_items = _sort_stac_items(items=presigned_stac_items, stac_ids=sort_by)

        # no filter
        if not stac_ids:
            return [item["assets"] for item in presigned_stac_items] if assets_only else presigned_stac_items

        stac_ids_set = set(stac_ids)
        sub_selection = [item for item in presigned_stac_items if item["id"] in stac_ids_set]
        return [item["assets"] for item in sub_selection] if assets_only else sub_selection

    def get_asset_bytesize(self, pre_signed_url: str) -> int:
        """get size in bytes of `pre_signed_url`"""
        return _get_asset_bytesize(pre_signed_url)

    # DOWNLOAD
    def download_asset(
        self,
        pre_signed_url: str,
        local_path: Union[Path, str] = None,
        override: bool = False,
        show_progress: bool = False,
    ) -> Path:
        """
        downloads a presigned asset url to disk

        Args:
            pre_signed_url: presigned asset url, see :py:meth:`get_presigned_assets`
            local_path: local output path - file is written to OS's temp dir if not provided
            override: override already existing `local_path`
            show_progress: show download status progressbar
        """
        dl_request = DownloadRequest(
            url=pre_signed_url,
            local_path=local_path,  # type: ignore
            asset_key="asset",
        )
        return _perform_download(
            download_requests=[dl_request],
            override=override,
            threaded=False,
            show_progress=show_progress,
        )["asset"]

    def download_products(
        self,
        assets_presigned: Optional[List[Dict[str, Any]]] = None,
        order_id: Optional[str] = None,
        tasking_request_id: Optional[str] = None,
        collect_id: Optional[str] = None,
        local_dir: Union[Path, str] = Path(tempfile.gettempdir()),
        include: Union[List[str], str] = None,
        exclude: Union[List[str], str] = None,
        override: bool = False,
        threaded: bool = True,
        show_progress: bool = False,
        separate_dirs: bool = True,
        product_types: List[str] = None,
    ) -> Dict[str, Dict[str, Path]]:
        """
        download all assets of multiple products

        Args:
            assets_presigned: mapping of presigned assets of multiple products, see :py:meth:`get_presigned_assets`
            order_id: optionally provide `order_id` instead of `assets_presigned`, see :py:meth:`submit_order`
            tasking_request_id: tasking request UUID of the task request you wish to download all associated products for
            collect_id: collect UUID you wish to download all associated products for

                    NOTE: Precedence order (high to low)
                      1. assets_presigned
                      2. order_id
                      3. tasking_request_id
                      4. collect_id

                    Meaning e.g. assets_presigned takes precedence over order_id, ...

            local_dir: local directory where assets are saved to, tempdir if not provided
            include: white-listing, which assets should be included, e.g. ["HH"] => only download HH asset
            exclude: black-listing, which assets should be excluded, e.g. ["HH", "thumbnail"] => download ALL except HH and thumbnail assets

                     NOTE: explicit DENY overrides explicit ALLOW

                     asset choices:
                        * 'HH', 'VV', 'raster', 'metadata', 'thumbnail' (external) - raster == 'HH' || 'VV'
                        * 'log', 'profile', 'stats', 'stats_plots' (internal)

            override: override already existing
            threaded: download assets of product in multiple threads
            show_progress: show download status progressbar
            separate_dirs: set to True in order to save the respective product assets into products directories, i.e.
                                /tmp/<stac_id_1>/<stac_id_1>.tif
                                /tmp/<stac_id_2>/<stac_id_2>.tif
                                ...
                            set to False in order to the respective product assets directly into the provided `local_dir`, i.e.
                               /tmp/<stac_id_1>.tif
                               /tmp/<stac_id_2>.tif
                               ...
            product_types: filter by product type, e.g. ["SLC", "GEO"]

        Returns:
            Dict[str, Dict[str, Path]]: Local paths of downloaded files keyed by STAC id and asset type, e.g.

            .. highlight:: python
            .. code-block:: python

                {
                    "stac_id_1": {
                        "<asset_type>": <path-to-asset>,
                        ...
                    }
                }
        """
        local_dir = Path(local_dir)

        one_of_required = (assets_presigned, order_id, tasking_request_id, collect_id)

        if not any(map(bool, one_of_required)):
            raise ValueError("please provide one of assets_presigned, order_id, tasking_request_id or collect_id")

        product_types = _validate_and_filter_product_types(product_types)
        include = _validate_and_filter_asset_types(include)
        exclude = _validate_and_filter_asset_types(exclude)

        if not assets_presigned:
            assets_presigned = self._resolve_assets_presigned(order_id, tasking_request_id, collect_id, product_types)

        len_assets_presigned = len(assets_presigned)
        suffix = "s" if len_assets_presigned > 1 else ""

        # filter product_type
        if product_types:
            assets_presigned = _filter_assets_by_product_types(assets_presigned, product_types)
        logger.info(f"downloading {len_assets_presigned} product{suffix}")

        # gather download requests
        download_requests = []
        by_stac_id = {}
        for cur_assets in assets_presigned:
            cur_download_requests = _gather_download_requests(cur_assets, local_dir, include, exclude, separate_dirs)
            by_stac_id[cur_download_requests[0].stac_id] = {
                cur.asset_key: cur.local_path for cur in cur_download_requests
            }
            download_requests.extend(cur_download_requests)

        if not download_requests:
            logger.warning("Nothing to download")
            return by_stac_id  # type: ignore

        # download
        _perform_download(
            download_requests=download_requests,
            override=override,
            threaded=threaded,
            show_progress=show_progress,
        )
        return by_stac_id  # type: ignore

    def _resolve_assets_presigned(
        self,
        order_id: Optional[str] = None,
        tasking_request_id: Optional[str] = None,
        collect_id: Optional[str] = None,
        product_types: List[str] = None,
    ) -> List[Dict[str, Any]]:

        stac_ids = None

        # 1 - resolve assets_presigned from order_id
        if order_id:
            _validate_uuid(order_id)
        else:
            # 2 - submit order for tasking_request_id
            if tasking_request_id:
                _validate_uuid(tasking_request_id)
                order_id, stac_ids = self._order_products_for_task(tasking_request_id, product_types)  # type: ignore
            # 3 - submit order for collect_id
            else:
                _validate_uuid(collect_id)
                order_id, stac_ids = self._order_products_for_collect_ids(
                    collect_ids=[collect_id], product_types=product_types  # type: ignore
                )

        return self.get_presigned_assets(order_id, stac_ids)  # type: ignore

    def _order_products_for_task(
        self, tasking_request_id: str, product_types: List[str] = None
    ) -> Tuple[str, List[str]]:
        """
        order all products associated with a tasking request

        Args:
            tasking_request_id: tasking request UUID you wish to order all associated products for
        """
        # gather up all collect IDs associated of this task
        collect_ids = [coll["collectId"] for coll in self.get_collects_for_task(tasking_request_id)]
        return self._order_products_for_collect_ids(collect_ids, product_types)

    def _order_products_for_collect_ids(
        self, collect_ids: List[str], product_types: List[str] = None
    ) -> Tuple[str, List[str]]:
        search_kwargs = dict(
            collect_id__in=collect_ids,
        )
        if product_types:
            search_kwargs["product_type__in"] = product_types

        result = self.search(**search_kwargs)
        if not result:
            logger.warning("No STAC items found ... aborting")
            sys.exit(0)

        order_id = self.submit_order(items=result, omit_search=True, check_active_orders=True)
        return order_id, result.stac_ids

    def download_product(
        self,
        assets_presigned: Optional[Dict[str, Any]] = None,
        order_id: Optional[str] = None,
        local_dir: Union[Path, str] = Path(tempfile.gettempdir()),
        include: Union[List[str], str] = None,
        exclude: Union[List[str], str] = None,
        override: bool = False,
        threaded: bool = True,
        show_progress: bool = False,
    ) -> Dict[str, Path]:
        """
        download all assets of a product

        Args:
            assets_presigned: mapping of presigned assets of multiple products, see :py:meth:`get_presigned_assets`
            order_id: optionally provide `order_id` instead of `assets_presigned`, see :py:meth:`submit_order`

                NOTE: Precedence order (high to low)
                  1. assets_presigned
                  2. order_id

            local_dir: local directory where assets are saved to, tempdir if not provided
            include: white-listing, which assets should be included, e.g. ["HH"] => only download HH asset
            exclude: black-listing, which assets should be excluded, e.g. ["HH", "thumbnail"] => download ALL except HH and thumbnail assets
                     NOTE: explicit DENY overrides explicit ALLOW

                     asset choices:
                        * 'HH', 'VV', 'raster', 'metadata', 'thumbnail' (external)
                           Note: raster == 'HH' || 'VV'
                        * 'log', 'profile', 'stats', 'stats_plots' (internal accessible only)

            override: override already existing
            threaded: download assets of product in multiple threads
            show_progress: show download status progressbar

        Returns:
            Dict[str, Path]: Local paths of downloaded files keyed by asset type, e.g.

            .. highlight:: python
            .. code-block:: python

                {
                    "<asset_type>": <path-to-asset>,
                    ...
                }
        """
        if not assets_presigned and not order_id:
            raise ValueError("please provide either assets_presigned or order_id")

        if not assets_presigned:
            _validate_uuid(order_id)
            assets_presigned = self._get_first_presigned_from_order(order_id)

        include = _validate_and_filter_asset_types(include)
        exclude = _validate_and_filter_asset_types(exclude)
        download_requests = _gather_download_requests(assets_presigned, local_dir, include, exclude)  # type: ignore

        if not download_requests:
            logger.warning("Nothing to download")
            return {}

        return _perform_download(
            download_requests=download_requests,
            override=override,
            threaded=threaded,
            show_progress=show_progress,
        )

    @no_type_check
    def _get_first_presigned_from_order(self, order_id: str) -> Dict[str, Any]:
        assets_presigned = self.get_presigned_assets(order_id)
        if len(assets_presigned) > 1:
            logger.warning(
                f"order {order_id} contains {len(assets_presigned)} products - using first one ({_derive_stac_id(assets_presigned)})"
            )

        return assets_presigned[0]

    # SEARCH
    def search(self, **kwargs) -> SearchResult:
        """
        paginated search for up to 500 matches (if no bigger limit specified)

        Find more information at https://docs.capellaspace.com/accessing-data/searching-for-data

        supported search filters:

         • bbox: List[float, float, float, float], e.g. [12.35, 41.78, 12.61, 42]
         • billable_area: Billable Area in m^2
         • center_frequency: Union[int, float], Center Frequency (GHz)
         • collections: List[str], e.g. ["capella-open-data"]
         • collect_id: str, capella internal collect-uuid, e.g. '78616ccc-0436-4dc2-adc8-b0a1e316b095'
         • constellation: str, e.g. "capella"
         • datetime: str, e.g. "2020-02-12T00:00:00Z"
         • frequency_band: str, Frequency band, one of "P", "L", "S", "C", "X", "Ku", "K", "Ka"
         • ids: List[str], e.g. `["CAPELLA_C02_SP_GEO_HH_20201109060434_20201109060437"]`
         • intersects: geometry component of the GeoJSON, e.g. {'type': 'Point', 'coordinates': [-113.1, 51.1]}
         • incidence_angle: Union[int, float], Center incidence angle, between 0 and 90
         • instruments: List[str], leveraged instruments, e.g. ["capella-radar-5"]
         • instrument_mode: str, Instrument mode, one of "spotlight", "stripmap", "sliding_spotlight"
         • limit: int, default: 500
         • look_angle: Union[int, float], e.g. 28.4
         • looks_azimuth: int, e.g. 5
         • looks_equivalent_number: int, Equivalent number of looks (ENL), e.g. 3
         • looks_range: int, e.g. 5
         • observation_direction: str, Antenna pointing direction, one of "right", "left"
         • orbit_state: str, Orbit State, one of "ascending", "descending"
         • orbital_plane: int, Orbital Plane, inclination angle of orbit
         • pixel_spacing_azimuth: Union[int, float], Pixel spacing azimuth (m), e.g. 0.5
         • pixel_spacing_range: Union[int, float], Pixel spacing range (m), e.g. 0.5
         • platform: str, e.g. "capella-2"
         • polarizations: str, one of "HH", "VV", "HV", "VH"
         • product_category: str, one of "standard", "custom", "extended"
         • product_type: str, one of "SLC", "GEO"
         • resolution_azimuth: float, Resolution azimuth (m), e.g. 0.5
         • resolution_ground_range: float, Resolution ground range (m), e.g. 0.5
         • resolution_range: float, Resolution range (m), e.g. 0.5
         • squint_angle: float, Squint angle, e.g. 30.1

        supported operations:
         • eq: equality search
         • in: within group
         • gt: greater than
         • gte: greater than equal
         • lt: lower than
         • lte: lower than equal

        sorting:
        • sortby: List[str] - must be supported fields, e.g. ["+datetime"]

        Returns:
            List[Dict[str, Any]]: STAC items matched
        """
        search = StacSearch(session=self._sesh, **kwargs)
        return search.fetch_all()


def _get_non_expired_orders(session: CapellaConsoleSession) -> List[Dict[str, Any]]:
    params = {"customerId": session.customer_id}
    res = session.get("/orders", params=params)

    all_orders = res.json()

    ordered_by_exp_date = sorted(all_orders, key=lambda x: x["expirationDate"])
    now = datetime.utcnow()

    active_orders = []
    while ordered_by_exp_date:
        cur = ordered_by_exp_date.pop()
        cur_exp_date = dateutil.parser.parse(cur["expirationDate"], ignoretz=True)
        if cur_exp_date < now:
            break
        active_orders.append(cur)

    return active_orders



import base64
from enum import Enum
from getpass import getpass

from typing import Optional, Tuple

import httpx

from capella_console_client.config import DEFAULT_TIMEOUT, CONSOLE_API_URL
from capella_console_client.hooks import log_on_4xx_5xx, translate_error_to_exception
from capella_console_client.logconf import logger
from capella_console_client.exceptions import (
    CapellaConsoleClientError,
    AuthenticationError,
    INVALID_TOKEN_ERROR_CODE,
    NoRefreshTokenError,
)
from capella_console_client.version import __version__


class AuthMethod(Enum):
    BASIC = 1  # email/ password
    TOKEN = 2  # JWT token


class CapellaConsoleSession(httpx.Client):
    def __init__(self, *args, **kwargs):
        verbose = kwargs.pop("verbose", False)
        search_url = kwargs.pop("search_url", None)
        event_hooks = [translate_error_to_exception]
        if verbose:
            event_hooks.insert(0, log_on_4xx_5xx)

        super(CapellaConsoleSession, self).__init__(
            *args,
            event_hooks={"response": event_hooks},
            timeout=DEFAULT_TIMEOUT,
            headers={
                "Content-Type": "application/json; charset=utf-8",
                "User-Agent": f"capella-console-client/{__version__}",
            },
            **kwargs,
        )
        self.customer_id = None
        self.organization_id = None

        self.search_url = search_url if search_url is not None else f"{self.base_url}/catalog/search"
        self._refresh_token = None

    def authenticate(
        self,
        email: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        no_token_check: bool = False,
    ) -> None:
        try:
            basic_auth_provided = bool(email) and bool(password)
            if not basic_auth_provided and not bool(token):
                email, password = self._prompt_user_creds(email, password)  # type: ignore

            auth_method = self._get_auth_method(email, password, token)
            if auth_method == AuthMethod.BASIC:
                self._basic_auth(email, password)  # type: ignore
            elif auth_method == AuthMethod.TOKEN:
                self._token_auth_check(token, no_token_check)  # type: ignore
        except httpx.HTTPStatusError:
            message = {
                AuthMethod.BASIC: "please check your credentials",
                AuthMethod.TOKEN: "provided token invalid",
            }[auth_method]

            raise AuthenticationError(
                f"Unable to authenticate with {self.base_url} ({auth_method}) - {message}"
            ) from None

        suffix = f"({self.base_url})" if self.base_url != CONSOLE_API_URL else ""
        if auth_method == AuthMethod.TOKEN and no_token_check:
            logger.info(f"successfully authenticated {suffix}")
        else:
            logger.info(f"successfully authenticated as {self.email} {suffix}")

    def _prompt_user_creds(self, email: str, password: str) -> Tuple[str, str]:
        """user credentials on console.capellaspace.com"""
        if not email:
            email = input("user on console.capellaspace.com (user@email.com): ").strip()
        if not password:
            password = getpass("password: ").strip()
        return (email, password)

    def _get_auth_method(self, email: Optional[str], password: Optional[str], token: Optional[str]) -> AuthMethod:
        basic_auth_provided = bool(email) and bool(password)
        has_token = bool(token)

        if not has_token and not basic_auth_provided:
            raise ValueError("please provide either email and password or token")

        if has_token and basic_auth_provided:
            logger.info("both token and email/ password provided ... using email/ password for authentication")

        auth_method = AuthMethod.BASIC
        if not basic_auth_provided:
            auth_method = AuthMethod.TOKEN
        return auth_method

    def _basic_auth(self, email: str, password: str):
        """
        authenticate with Console API

        returns jwt access token
        """
        basic_token = base64.b64encode(f"{email}:{password}".encode()).decode("utf-8")
        resp = self.post("/token", headers={"Authorization": f"Basic {basic_token}"})
        resp.raise_for_status()
        response_body = resp.json()

        self._set_auth_header(response_body["accessToken"])
        self._refresh_token = response_body["refreshToken"]
        self._cache_user_info()

    def _set_auth_header(self, token: str):
        token = token.strip()
        if not token.startswith("Bearer"):
            token = f"Bearer {token}"

        self.headers["Authorization"] = token

    def _cache_user_info(self):
        """cache customer_id and organization_id - serves as test for successful auth"""
        resp = self.get("/user")
        resp.raise_for_status()

        con = resp.json()
        self.customer_id = con["id"]
        self.organization_id = con["organizationId"]
        self.email = con["email"]

    def _token_auth_check(self, token: str, no_token_check: bool):
        self._set_auth_header(token)
        if not no_token_check:
            self._cache_user_info()

    def send(self, *fct_args, **kwargs):
        """wrap httpx.Client.send for auto token_refresh"""
        try:
            ret = super().send(*fct_args, **kwargs)
        except AuthenticationError as e:
            # safeguard in case AuthenticationError get's improperly re-used
            if e.code != INVALID_TOKEN_ERROR_CODE:
                raise e

            self.perform_token_refresh()

            # retry request
            orig_request = fct_args[0]
            orig_request.headers["authorization"] = self.headers["authorization"]
            ret = super().send(*fct_args, **kwargs)
        return ret

    def perform_token_refresh(self):
        if not self._refresh_token:
            raise NoRefreshTokenError("No refresh token found") from None

        resp = self.post("/token/refresh", json={"refreshToken": self._refresh_token})
        con = resp.json()
        self._set_auth_header(con["accessToken"])
        if con["refreshToken"] != self._refresh_token:
            self._refresh_token = con["refreshToken"]
        logger.info("successfully refreshed access token")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['capella_console_client',
 'capella_console_client.cli',
 'capella_console_client.cli.user_searches']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.23.0,<0.24.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'retrying>=1.3.3,<2.0.0',
 'rich>=12.5.1,<13.0.0']

extras_require = \
{'docs': ['Sphinx>=5.1.1,<6.0.0',
          'sphinx-rtd-theme>=1.0.0,<2.0.0',
          'sphinx-autobuild>=2021.3.14,<2022.0.0',
          'sphinx-copybutton>=0.5.0,<0.6.0',
          'sphinx-autodoc-typehints>=1.19.2,<2.0.0'],
 'wizard': ['typer>=0.6.1,<0.7.0',
            'questionary>=1.10.0,<2.0.0',
            'tabulate>=0.8.10,<0.9.0']}

entry_points = \
{'console_scripts': ['capella-console-wizard = '
                     'capella_console_client.cli.wizard:main']}

setup_kwargs = {
    'name': 'capella-console-client',
    'version': '0.9.1',
    'description': 'Python SDK for api.capellaspace.com (search, order, download)',
    'long_description': '# 🛰️ capella-console-client 🐐\n\n[![Version](https://img.shields.io/pypi/v/capella-console-client.svg)](https://pypi.org/project/capella-console-client/)\n[![License](https://img.shields.io/pypi/l/capella-console-client.svg)](#)\n[![CI](https://github.com/capellaspace/console-client/workflows/CI/badge.svg)](#)\n[![Coverage](https://coveralls.io/repos/github/capellaspace/console-client/badge.svg?branch=main)](https://coveralls.io/github/capellaspace/console-client)\n[![Supported Python Versions](https://img.shields.io/pypi/pyversions/capella-console-client.svg)](https://pypi.org/project/capella-console-client/)\n[![Documentation](https://readthedocs.org/projects/capella-console-client/badge/?version=main)](https://capella-console-client.readthedocs.io)\n\nPython SDK for api.capellaspace.com (search, order, download)\n\n\n## Installation\n\n```bash\npip install capella-console-client\n```\n\n## Requirements\n\n* python >= 3.7\n* `capella-console-client` requires an active account on [console.capellaspace.com](https://console.capellaspace.com/). Sign up for an account at [https://www.capellaspace.com/community/](https://www.capellaspace.com/community/).\n\n\n## Usage\n\n![Quickstart](docs/images/quickstart.gif)\n\n```python\nfrom capella_console_client import CapellaConsoleClient\n\n# you will be prompted for console user (user@email.com)/ password before authenticating\nclient = CapellaConsoleClient(\n    verbose=True\n)\n\n# search for 2 open-data products\nstac_items = client.search(\n    instrument_mode="spotlight",\n    product_type__in=["SLC", "GEO"],\n    collections=["capella-open-data"],\n    limit=2\n)\n\n# order\norder_id = client.submit_order(items=stac_items, omit_search=True)\n\n# download\nproduct_paths = client.download_products(\n    order_id=order_id, \n    local_dir=\'/tmp\',\n    show_progress=True\n)\n```\n\n\n## Documentation\n\nThe documentation for `capella-console-client` can be found [here](https://capella-console-client.readthedocs.io).\n\n## 🧙\u200d capella-console-wizard 🧙\u200d♂️\nstarting with `capella-console-client>=0.8.0` the SDK ships with an interactive wizard-like CLI: `capella-console-wizard` \n\n### Installation\n```\npip install capella-console-client[wizard]\n```\n\n### Usage\n```\ncapella-console-wizard --help\n```\n\nsee \n\n\n## Support\n\nPlease [open an issue](https://github.com/capellaspace/console-client/issues/new)\nwith enough information for us to reproduce your problem.\nA [minimal, reproducible example](https://stackoverflow.com/help/minimal-reproducible-example)\nwould be very helpful.\n\n## Contributing\n\nContributions are very much welcomed and appreciated. See [how to contribute](https://capella-console-client.readthedocs.io/en/main/pages/contributors.html) for more information.\n\n\n## License\n• Licensed under the [MIT License](https://github.com/capellaspace/console-client/blob/master/LICENSE) • Copyright 2022 • Capella Space •\n',
    'author': 'Thomas Beyer',
    'author_email': 'thomas.beyer@capellaspace.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/capellaspace/console-client',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
from typing import List
from dataclasses import dataclass

import httpx

from capella_console_client.exceptions import (
    CapellaConsoleClientError,
    handle_error_response,
    NON_RETRYABLE_ERROR_CODES,
)

logger = logging.getLogger()


@dataclass
class RequestMeta:
    method: str
    url: httpx.URL


# do not log out the following requests
SILENCE_REQUESTS: List[RequestMeta] = []


def translate_error_to_exception(response):
    if response.status_code >= 500:
        handle_error_response(response)


def log_on_4xx_5xx(response):
    try:
        response.raise_for_status()
    except Exception:
        request = response.request
        cur = RequestMeta(request.method, request.url)
        if cur in SILENCE_REQUESTS:
            return

        logger.error(
            f"Request: {request.method} {request.url} - Status {response.status_code} - Response: {response.json()}"
        )
        return True


def retry_if_http_status_error(exception):
    """Return upon httpx.HTTPStatusError"""
    if getattr(exception, "code", None) in NON_RETRYABLE_ERROR_CODES:
        return False
    return isinstance(exception, CapellaConsoleClientError)


def retry_if_httpx_status_error(exception):
    return isinstance(exception, httpx.HTTPStatusError)


def log_attempt_delay(attempts, delay):
    logger.info(f"Attempt #{attempts}, retrying in {delay} ms")
    return delay
